#
# Copyright 2019-2021 Lars Pastewka
#           2020-2021 Michael Röttger
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import json

import numpy as np

from numpyencoder import NumpyEncoder

from ..Exceptions import MetadataAlreadyFixedByFile
from ..HeightContainer import UniformTopographyInterface, NonuniformLineScanInterface
from ..NonuniformLineScan import NonuniformLineScan
from ..UniformLineScanAndTopography import Topography, UniformLineScan
from ..Support.UnitConversion import mangle_length_unit_utf8, mangle_length_unit_ascii

from .Reader import ReaderBase, ChannelInfo

# We run serial I/O through scipy. This has several advantages:
# 1) lightweight, 2) can handle streams
from scipy.io import netcdf_file

format_to_scipy_version = {
    'NETCDF3_CLASSIC': 1,
    'NETCDF3_64BIT_OFFSET': 2
}


# we subclass the netcdf_file class such it
# is not closed by garbage collector in case of file streams
class _SpecialNetCDFFile(netcdf_file):
    def close(self):
        if self.filename != 'None':
            super().close()
        else:
            self.flush()
        # we want to disable calling closing the file
        # in case of file streams

    __del__ = close


class NCReader(ReaderBase):
    _format = 'nc'
    _name = 'Network Common Data Format (NetCDF)'

    _description = '''
This reader reads topography data contained in a
[NetCDF](https://www.unidata.ucar.edu/software/netcdf/) container. The
reader looks for a variable named `heights` containing a two-dimensional
array that is interpreted as height information. The respective dimensions are
named `x` and `y`.

The reader additionally looks for two (optional) variables `x` and `y` that
contain the x- and y-coordinates of the first and second index of the height
arrays. The attribute `length` of `x` and `y` must contain the physical size
in the respective direction. The optional attribute `length_unit` of these
variables describes the physical unit. The optional additional attribute
`periodic` indicates whether the direction contains periodic data. If
`periodic` is missing, the reader interprets the data as non-periodic.

If an additional variable `mask` is present, this variable is used to mark
undefined points.

An example file layout (output of `ncdump -h`) containing a topography map
with 128 x 128 pixels looks like this:
```
netcdf test_nc_file {
dimensions:
    x = 128 ;
    y = 128 ;
variables:
    double x(x) ;
        x:length = 3 ;
        x:periodic = 1 ;
        x:unit = "um" ;
    double y(y) ;
        y:length = 3 ;
        y:periodic = 1 ;
        y:unit = "um" ;
    double heights(x, y) ;
        heights:unit = "um" ;
}
```
The following code snippets reads the file and displays the topography data as
a two-dimensional color map in Python:
```python
import numpy as np
import matplotlib.pyplot as plt
from scipy.io.netcdf import netcdf_file

nc = netcdf_file('test.nc')
heights = np.array(nc.variables['heights'][...])
x = np.array(nc.variables['x'][...])
y = np.array(nc.variables['y'][...])
unit = nc.variables['x'].unit

plt.figure()
plt.subplot(aspect=1)
plt.pcolormesh(x, y, heights.T)

plt.show()
```
'''

    def __init__(self, fobj, communicator=None):
        self._nc = None
        self._var_kwargs = {}  # Keywords arguments to attach to variable creation
        if communicator is not None and communicator.size > 1:
            # For parallel I/O we need netCDF4
            from netCDF4 import Dataset
            self._nc = Dataset(fobj, 'r', parallel=True, comm=communicator, format='NETCDF3_64BIT_OFFSET')
        else:
            # We need to check magic ourselves if this is a stream because
            # netcdf_file may close the stream in case of error, probably
            # because of garbage collection
            if hasattr(fobj, 'seek'):
                magic = fobj.read(3)
                if not magic == b'CDF':
                    raise TypeError('File or stream is not a valid NetCDF 3 file')
                # we must rewind the file
                fobj.seek(0)

            self._nc = _SpecialNetCDFFile(fobj, 'r', maskandscale=True)

        self._communicator = communicator
        self._n_dim = self._nc.dimensions['n'] if 'n' in self._nc.dimensions else None
        self._x_dim = self._nc.dimensions['x'] if 'x' in self._nc.dimensions else None
        self._y_dim = self._nc.dimensions['y'] if 'y' in self._nc.dimensions else None
        self._x_var = self._nc.variables['x'] if 'x' in self._nc.variables else None
        self._y_var = self._nc.variables['y'] if 'y' in self._nc.variables else None
        self._heights_var = self._nc.variables['heights']
        self._mask_var = self._nc.variables['mask'] if 'mask' in self._nc.variables else None

        # The following information may be missing from the NetCDF file
        self._physical_sizes = None
        self._periodic = False
        self._unit = None
        self._info = {}

        # Deserialize info dictionary, if present
        try:
            self._info = json.loads(self._nc.json)
        except AttributeError:
            pass

        # Determine physical sizes of topography
        if self._n_dim is not None:
            # This is a nonuniform topography
            pass
        else:
            # This is a uniform topography...
            if self._y_var is not None:
                # ...and it is 2D
                try:
                    self._physical_sizes = (self._x_var.length, self._y_var.length)
                except AttributeError:
                    pass
            else:
                # ...and it is 1D (a line scan)
                try:
                    self._physical_sizes = (self._x_var.length,)
                except AttributeError:
                    pass

        # Determine if the topography is periodic
        try:
            self._periodic = self._x_var.periodic != 0
        except AttributeError:
            pass

        # Determine unit of topography
        try:
            self._unit = mangle_length_unit_utf8(self._x_var.unit)
        except AttributeError:
            pass

        # Check unit consistency
        if self._y_var is not None:
            try:
                y_unit = mangle_length_unit_utf8(self._y_var.unit)
            except AttributeError:
                y_unit = None
            if y_unit is not None and y_unit != self._unit:
                raise ValueError('Units for x- and y-coordinate in NetCDF file differ.')

        try:
            heights_unit = mangle_length_unit_utf8(self._heights_var.unit)
        except AttributeError:
            heights_unit = None
        if heights_unit is not None and heights_unit != self._unit:
            raise ValueError('Units for x-coordinate and heights in NetCDF file differ.')

    def __del__(self):
        self.close()

    def close(self):
        if self._nc is not None:
            if self._x_var is not None:
                del self._x_var
            if self._y_var is not None:
                del self._y_var
            if self._heights_var is not None:
                del self._heights_var

            del self._nc
            self._nc = None

    @property
    def channels(self):
        if self._x_dim is not None:
            # This is a uniform topography
            uniform = True
            try:
                # netCDF4
                nx = self._x_dim.size
            except AttributeError:
                # scipy.io.netcdf_file
                nx = self._x_dim
            nb_grid_pts = (nx,)
            if self._y_dim is not None:
                # This is a 2D topography
                try:
                    # netCDF4
                    ny = self._y_dim.size
                except AttributeError:
                    # scipy.io.netcdf_file
                    ny = self._y_dim
                nb_grid_pts = (nx, ny)
        else:
            # This is a nonuniform line scan
            uniform = False
            try:
                # netCDF4
                nx = self._x_dim.size
            except AttributeError:
                # scipy.io.netcdf_file
                nx = self._x_dim
            nb_grid_pts = (nx,)
        return [ChannelInfo(self, 0,
                            name='Default',
                            dim=len(nb_grid_pts),
                            nb_grid_pts=nb_grid_pts,
                            physical_sizes=self._physical_sizes,
                            periodic=self._periodic,
                            uniform=uniform,
                            undefined_data=self._mask_var is not None,
                            unit=self._unit,
                            info=self._info)]

    def topography(self, channel_index=None, physical_sizes=None, height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None, nb_subdomain_grid_pts=None):
        if channel_index is not None and channel_index != 0:
            raise ValueError('`channel_index` must be None or 0.')

        if unit is not None:
            if self._unit is not None:
                raise MetadataAlreadyFixedByFile('unit')
        else:
            unit = self._unit

        _info = self._info.copy()
        _info.update(info)

        heights = self._heights_var[...]
        if self._mask_var is not None:
            heights = np.ma.masked_where(self._mask_var[...], heights)
        if not np.ma.is_masked(heights):
            # If is it not masked, make sure it becomes a proper numpy array
            heights = np.array(heights)
        heights = heights.copy()

        if subdomain_locations is None and nb_subdomain_grid_pts is None:
            if self._x_dim is not None:
                physical_sizes = self._check_physical_sizes(physical_sizes,
                                                            self._physical_sizes)
                # This is a uniform topography...
                if self._y_dim is not None:
                    # ...and it is 2D
                    topo = Topography(heights, physical_sizes,
                                      periodic=self._periodic if periodic is None else periodic,
                                      unit=unit, info=_info)
                else:
                    # .. and it is 1D
                    topo = UniformLineScan(heights, physical_sizes,
                                           periodic=self._periodic if periodic is None else periodic,
                                           unit=unit, info=_info)
            else:
                if physical_sizes is not None:
                    raise ValueError('You cannot specify physical sizes for a nonuniform topography.')

                if self._periodic:
                    raise ValueError('NetCDF file indicates periodicity, but this is a nonuniform line scan which '
                                     'cannot be periodic.')

                # This is a nonuniform line scan
                topo = NonuniformLineScan(self._x_var[...].copy(), heights,
                                          unit=unit, info=_info)
        else:
            if self._y_dim is None:
                raise ValueError('Parallel reading only works for topographies, not line scans.')

            physical_sizes = self._check_physical_sizes(physical_sizes,
                                                        self._physical_sizes)

            topo = Topography(heights, physical_sizes,
                              periodic=self._periodic if periodic is None else periodic,
                              decomposition='domain',
                              subdomain_locations=subdomain_locations,
                              nb_subdomain_grid_pts=nb_subdomain_grid_pts,
                              communicator=self.communicator,
                              unit=unit,
                              info=_info)

        if height_scale_factor is not None:
            if 'unit' in _info:
                raise MetadataAlreadyFixedByFile('height_scale_factor',
                                                 'The height_scale_factor cannot be given for NC files '
                                                 'if units are present in the file.')
                # See https://github.com/ContactEngineering/SurfaceTopography/pull/99#discussion_r648364687
            topo = topo.scale(height_scale_factor)

        return topo

    @property
    def communicator(self):
        return self._communicator

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__


def write_nc_uniform(topography, fobj, format='NETCDF3_64BIT_OFFSET'):
    """
    Write topography into a NetCDF file.

    Parameters
    ----------
    topography : :obj:`SurfaceTopography`
        The topography to write to disk.
    fobj : str or stream
        Name of the NetCDF file or file stream
    format : str
        NetCDF file format. Default is 'NETCDF3_64BIT_OFFSET'.
    """
    if topography.communicator is not None and topography.communicator.size > 1:
        from netCDF4 import Dataset, default_fillvals
        kwargs = dict(format=format,
                      parallel=topography.is_domain_decomposed,
                      comm=topography.communicator)
        var_kwargs = {'fill_value': default_fillvals['f8']}
    else:
        Dataset = _SpecialNetCDFFile
        kwargs = dict(version=format_to_scipy_version[format],
                      maskandscale=True)
        var_kwargs = {}
    if not topography.is_domain_decomposed and \
            topography.communicator.rank > 1:
        return
    with Dataset(fobj, 'w', **kwargs) as nc:
        # Serialize info dictionary as JSON and write to NetCDF file
        info = topography.info.copy()
        try:
            # We store the unit information separately
            del info['unit']
        except KeyError:
            pass
        if info != {}:
            nc.json = json.dumps(info, ensure_ascii=True, cls=NumpyEncoder)

        # Create dimensions and heights variable
        try:
            nx, ny = topography.nb_grid_pts
        except ValueError:
            nx, = topography.nb_grid_pts

        nc.createDimension('x', nx)
        if topography.dim > 1:
            nc.createDimension('y', ny)
            spatial = ('x', 'y')
            heights_var = nc.createVariable('heights', 'f8', spatial, **var_kwargs)
        else:
            spatial = ('x',)
            heights_var = nc.createVariable('heights', 'f8', spatial, **var_kwargs)

        # Storing and retrieving masked arrays does not work reliably in
        # NetCDF. We need to store the mask explicitly.
        mask_var = None
        if topography.has_undefined_data:
            mask_var = nc.createVariable('mask', 'b', spatial, **var_kwargs)

        # Create variables for x- and y-positions, but only if physical_sizes
        # exist. (physical_sizes should always exist, but who knows...)
        if topography.physical_sizes is not None:
            x = topography.positions()
            try:
                sx, sy = topography.physical_sizes
                x, y = x
            except ValueError:
                sx, = topography.physical_sizes

            x_var = nc.createVariable('x', 'f8', ('x',), **var_kwargs)
            x_var.length = sx

            x_var.periodic = 1 if topography.is_periodic else 0
            if topography.unit is not None:
                # scipy.io.netcdf_file does not support UTF-8
                x_var.unit = mangle_length_unit_ascii(topography.unit)

            if topography.dim > 1:
                x_var[...] = np.arange(nx) / nx * sx

                y_var = nc.createVariable('y', 'f8', ('y',), **var_kwargs)

                y_var.length = sy
                y_var.periodic = 1 if topography.is_periodic else 0
                if topography.unit is not None:
                    # scipy.io.netcdf_file does not support UTF-8
                    y_var.unit = mangle_length_unit_ascii(topography.unit)
                y_var[...] = np.arange(ny) / ny * sy

        h = topography.heights()
        if topography.is_domain_decomposed:
            heights_var.set_collective(True)
            heights_var[topography.subdomain_slices] = h
            if mask_var is not None:
                mask_var[topography.subdomain_slices] = np.ma.getmask(h)
        else:
            heights_var[...] = h
            if mask_var is not None:
                mask_var[...] = np.ma.getmask(h)
        if topography.unit is not None:
            heights_var.unit = mangle_length_unit_ascii(topography.unit)


def write_nc_nonuniform(line_scan, fobj, format='NETCDF3_64BIT_OFFSET'):
    """
    Write nonuniform line scan into a NetCDF file.

    Parameters
    ----------
    line_scan : :obj:`SurfaceTopography`
        The topography to write to disk.
    fobj : str or stream
        Name of the NetCDF file or file stream
    format : str
        NetCDF file format. Default is 'NETCDF3_64BIT_OFFSET'.
    """
    if line_scan.communicator is not None and line_scan.communicator.size > 1:
        raise RuntimeError('Parallel writing is not supported for nonuniform line scans')

    with _SpecialNetCDFFile(fobj, 'w', version=format_to_scipy_version[format], maskandscale=True) as nc:
        # Serialize info dictionary as JSON and write to NetCDF file
        info = line_scan.info.copy()
        try:
            # We store the unit information separately
            del info['unit']
        except KeyError:
            pass
        if info != {}:
            nc.json = json.dumps(info)

        nx, = line_scan.nb_grid_pts

        nc.createDimension('n', nx)

        x_var = nc.createVariable('x', 'f8', ('n',))
        heights_var = nc.createVariable('heights', 'f8', ('n',))

        x_var[...] = line_scan.positions()
        if line_scan.unit is not None:
            # scipy.io.netcdf_file does not support UTF-8
            x_var.unit = mangle_length_unit_ascii(line_scan.unit)
        heights_var[...] = line_scan.heights()


# Register analysis functions from this module
UniformTopographyInterface.register_function('to_netcdf', write_nc_uniform)
NonuniformLineScanInterface.register_function('to_netcdf', write_nc_nonuniform)



#
# Copyright 2021 Michael Röttger
#           2020-2021 Lars Pastewka
#           2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import warnings
from functools import wraps


def deprecated(version=None, alternative=None):
    """Emit a deprecation warning for a function."""

    def deprecated_decorator(func):

        def _get_warn_str(version=version, alternative=alternative):
            if version is None:
                warn_str = f'Function/property `{func.__name__}` is deprecated.'
            else:
                warn_str = f'Function/property `{func.__name__}` was deprecated in SurfaceTopography {version}.'
            if alternative is not None:
                warn_str += f' The function/property `{alternative}` provides this functionality now.'
            return warn_str

        @wraps(func)
        def deprecated_func(*args, **kwargs):
            warnings.warn(_get_warn_str(version, alternative), DeprecationWarning)
            return func(*args, **kwargs)

        docstring = deprecated_func.__doc__ or ""
        docstring += "\n\n" + _get_warn_str(version, alternative)
        deprecated_func.__doc__ = docstring

        return deprecated_func

    return deprecated_decorator


class DeprecatedDictionary(dict):
    """A dictionary that raises a deprecation warning when accessing certain keys."""

    def __init__(self, *args, **kwargs):
        self._deprecated_keys = kwargs['deprecated_keys']
        del kwargs['deprecated_keys']
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        if key in self._deprecated_keys:
            warnings.warn(f"Dictionary key '{key}' has been deprecated.", DeprecationWarning)
        return super().__getitem__(key)

    def copy(self):
        return DeprecatedDictionary(self, deprecated_keys=self._deprecated_keys)



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# Reference information and implementations:
# https://sourceforge.net/p/gwyddion/code/HEAD/tree/trunk/gwyddion/modules/file/psia.c

import io

import numpy as np

from tiffile import TiffFile, TiffFileError

from .binary import decode
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo
from ..Exceptions import CorruptFile, FileFormatMismatch, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography
from ..Support.UnitConversion import get_unit_conversion_factor


class PSReader(ReaderBase):
    _format = 'ps'
    _name = 'Park Systems'
    _description = '''
TIFF-based file format of Park Systems instruments.
'''

    _MAGIC = 0x0E031301
    _VERSION1 = 0x1000001
    _VERSION2 = 0x1000002

    _TAG_MAGIC = 50432
    _TAG_VERSION = 50433
    _TAG_DATA = 50434
    _TAG_HEADER = 50435

    _IMAGE_TYPE_2D_MAPPED = 0
    _IMAGE_TYPE_LINE_PROFILE = 1
    _IMAGE_TYPE_SPECTROSCOPY = 2

    _DATA_TYPE_INT16 = 0
    _DATA_TYPE_INT32 = 1
    _DATA_TYPE_FLOAT = 2

    _header_structure = [
        ('image_type', 'I'),
        ('source_name', '32U'),
        ('image_mode', '8U'),
        ('lpf_strength', 'd'),
        ('auto_flatten', 'I'),
        ('ac_track', 'I'),
        ('nb_grid_pts_x', 'I'),
        ('nb_grid_pts_y', 'I'),
        ('angle', 'd'),
        ('sine_scan', 'I'),
        ('overscan_rate', 'd'),
        ('forward', 'I'),
        ('scan_up', 'I'),
        ('swap_xy', 'I'),
        ('physical_size_x', 'd'),
        ('physical_size_y', 'd'),
        ('xoff', 'd'),
        ('yoff', 'd'),
        ('scan_rate', 'd'),
        ('set_point', 'd'),
        ('set_point_unit', '8U'),
        ('tip_bias', 'd'),
        ('sample_bias', 'd'),
        ('data_gain', 'd'),
        ('data_scale_factor', 'd'),
        ('data_offset', 'd'),
        ('data_unit', '8U'),
        ('data_min', 'i'),
        ('data_max', 'i'),
        ('data_avg', 'i'),
        ('compression', 'I'),
        ('logscale', 'I'),
        ('square', 'I'),
    ]

    _version2_header_structure = [
        ('data_servo_gain', 'd'),
        ('data_scanner_range', 'd'),
        ('xy_voltage_mode', '8U'),
        ('data_voltage_mode', '8U'),
        ('xy_servo_mode', '8U'),
        ('data_type', 'I'),
        ('reserved1', 'I'),
        ('reserved2', 'I'),
        ('ncm_amplitude', 'd'),
        ('ncm_frequency', 'd'),
        ('cantilever_name', '16U'),
    ]

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self._file_path = file_path
        with OpenFromAny(self._file_path, 'rb') as f:
            try:
                with TiffFile(f) as t:
                    if len(t.pages) != 1:
                        raise FileFormatMismatch('More than one image in TIFF.')
                    p = t.pages[0]

                    # Check file magic and version information
                    if p.tags[self._TAG_MAGIC].value != self._MAGIC:
                        raise FileFormatMismatch('This is not a Park Systems TIFF.')
                    self._version = p.tags[self._TAG_VERSION].value
                    if self._version not in [self._VERSION1, self._VERSION2]:
                        raise CorruptFile('Only version 1 and 2 of Park Systems TIFFs are supported.')

                    # Parse header
                    header_data = p.tags[self._TAG_HEADER].value
                    if (self._version == self._VERSION1 and len(header_data) < 356) or (
                            self._version == self._VERSION2 and len(header_data) < 580):
                        raise CorruptFile('Header too short.')

                    header_stream = io.BytesIO(header_data)
                    self._header = decode(header_stream, self._header_structure, '<')

                    if self._version == self._VERSION2:
                        self._header.update(decode(header_stream, self._version2_header_structure, '<'))
                    else:
                        self._header['data_type'] = self._DATA_TYPE_INT16

                    if self._header['data_type'] not in [self._DATA_TYPE_INT16, self._DATA_TYPE_INT32,
                                                         self._DATA_TYPE_FLOAT]:
                        raise CorruptFile('Cannot handle data type {}'.format(self._header['data_type']))

                    self._nb_grid_pts = (self._header['nb_grid_pts_x'], self._header['nb_grid_pts_y'])
                    self._physical_sizes = (self._header['physical_size_x'], self._header['physical_size_y'])
                    self._unit = 'µm'

                    unit_conversion_factor = get_unit_conversion_factor(self._header['data_unit'], 'µm')
                    self._height_scale_factor = \
                        self._header['data_scale_factor'] * self._header['data_gain'] * unit_conversion_factor

                    self._info = {'raw_metadata': self._header}
            except TiffFileError:
                raise FileFormatMismatch('This is not a TIFF file, so it cannot be a Park Systems TIFF.')

    @property
    def channels(self):
        return [ChannelInfo(self,
                            0,  # channel index
                            name=self._header['source_name'],
                            dim=2,
                            nb_grid_pts=self._nb_grid_pts,
                            physical_sizes=self._physical_sizes,
                            height_scale_factor=self._height_scale_factor,
                            uniform=True,
                            unit=self._unit)]

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI parallelization.')

        if channel_index is None:
            channel_index = self._default_channel_index

        if channel_index != self._default_channel_index:
            raise RuntimeError(
                f'There is only a single channel. Channel index must be {self._default_channel_index}.')

        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        with OpenFromAny(self._file_path, 'rb') as f:
            with TiffFile(f) as t:
                # The data itself is not the image, but inside a tag
                raw_data = t.pages[0].tags[self._TAG_DATA].value
                t = self._header['data_type']
                dtype = np.int16 if t == self._DATA_TYPE_INT16 else \
                    np.int32 if t == self._DATA_TYPE_INT32 else np.single
                height_data = \
                    np.frombuffer(raw_data, dtype=dtype, count=np.prod(self._nb_grid_pts)).reshape(self._nb_grid_pts).T

        _info = self._info.copy()
        _info.update(info)

        topo = Topography(height_data,
                          self._physical_sizes,
                          unit=self._unit,
                          periodic=False if periodic is None else periodic,
                          info=_info)
        return topo.scale(self._height_scale_factor)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Analysis functions related to scanning-probe microscopy
"""

import numpy as np
import scipy

from ..HeightContainer import NonuniformLineScanInterface, UniformTopographyInterface
from ..Support import doi


@doi('10.1016/j.apsadv.2021.100190')
def scanning_probe_reliability_cutoff(self, tip_radius, safety_factor=1 / 2):
    """
    Estimate a length scale below which tip radius artifacts affect the
    measured data. See: 10.1016/j.apsadv.2021.100190

    Parameters
    ----------
    self : Topography or UniformLineScan
        Topogaphy or line scan.
    tip_radius : float
        Tip radius.
    safety_factor : float, optional
        Tip radius artifacts are expected to play a role for scales below
        which the minimum scale-dependent curvature drops below
        -safety_factor/radius. The `safety_factor` should be on the order of 1.
        In 10.1016/j.apsadv.2021.100190 a factor of 1/2 is estimated based
        on synthetic (simulated) data. (Default: 1/2)

    Returns
    -------
    reliability_cutoff : float
        Length scale below which data is affected by tip radius artifacts.
        Note that this is the length of the stencil for the second derivative
        used to compute the reliability cutoff.
    """
    target_curvature = safety_factor / tip_radius

    def objective(distance):
        d = self.derivative(n=2, distance=distance)
        if self.dim == 1:
            if len(d) == 0:
                return target_curvature
            return target_curvature + np.min(d)
        elif self.dim == 2:
            if len(d[0]) == 0:
                return target_curvature
            return target_curvature + np.min(d[0])
        else:
            raise ValueError(f'Cannot handle a {self.dim}-dimensional topography.')

    lower, upper = self.bandwidth()
    if objective(2 * lower) > 0:
        # Curvature is at lower end is smaller than tip curvature
        return None
    elif objective(upper / 2) < 0:
        # Curvature at upper end is larger than tip curvature;
        # None of the data is reliable
        return upper
    else:
        return scipy.optimize.brentq(objective,
                                     2 * lower, upper / 2,  # bounds
                                     xtol=1e-4)


UniformTopographyInterface.register_function('scanning_probe_reliability_cutoff', scanning_probe_reliability_cutoff)
NonuniformLineScanInterface.register_function('scanning_probe_reliability_cutoff', scanning_probe_reliability_cutoff)



#
# Copyright 2018-2020 Lars Pastewka
#           2019 Antoine Sanner
#           2019 Michael Röttger
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Bin for small common helper function and classes for nonuniform
topographies.
"""

import numpy as np

from ..HeightContainer import NonuniformLineScanInterface


def bandwidth(self):
    """
    Computes lower and upper bound of bandwidth, i.e. of the wavelengths or
    length scales occurring on a topography. The lower end of the bandwidth is
    given by the mean of the spacing of the individual points on the line
    scan. The upper bound is given by the overall length of the line scan.

    Returns
    -------
    lower_bound : float
        Lower bound of the bandwidth.
    upper_bound : float
        Upper bound of the bandwidth.
    """
    x = self.positions()
    lower_bound = np.mean(np.diff(x))
    upper_bound, = self.physical_sizes

    return lower_bound, upper_bound


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('bandwidth', bandwidth)



#
# Copyright 2020-2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


class SurfaceContainer(object):
    _functions = {}

    def __init__(self, topographies=[]):
        self._topographies = topographies

    def __len__(self):
        return len(self._topographies)

    def __getitem__(self, item):
        return self._topographies[item]

    def apply(self, name, *args, **kwargs):
        self._functions[name](self, *args, **kwargs)

    def __getattr__(self, name):
        if name in self._functions:
            def func(*args, **kwargs):
                return self._functions[name](self, *args, **kwargs)

            func.__doc__ = self._functions[name].__doc__
            return func
        else:
            raise AttributeError("Unkown attribute '{}' and no analysis or pipeline function of this name registered"
                                 "(class {}). Available functions: {}".format(name, self.__class__.__name__,
                                                                              ', '.join(self._functions.keys())))

    def __dir__(self):
        return sorted(super().__dir__() + [*self._functions])

    @classmethod
    def register_function(cls, name, function):
        cls._functions.update({name: function})



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Tracing bibliography through function calls.
"""

_default_dois = set()


class doi(object):
    """
    To add bibliographic information, simply decorate the function:

    ```
    @doi('10.1088/2051-672X/aa51f8')
    def power_spectrum(topography):
        ...
    ```

    The DOIs can be requested by passing a `doi` argument with an empty set to
    the function:

    ```
    dois = set()
    power_spectrum(topography, dois=dois)
    print(dois)  # Prints the relevant bibliographic information
    ```

    Note that the dois need to be evaluated directly after the function call.
    Any additional function calls will continue to populate the set with
    bibliography information.
    """

    dois = set()

    _n = 0

    def __init__(self, *args):
        self._add_these_dois = args

    def __call__(self, func):
        def func_with_doi(*args, **kwargs):
            if 'dois' in kwargs:
                # `dois` is present. We store a reference to this set that
                # will be updated in subsequent function calls.
                doi.dois = kwargs['dois']
                # Update this with the default set of DOIs
                doi.dois.update(_default_dois)
                del kwargs['dois']
                doi._n = 0
            doi.dois.update(self._add_these_dois)
            doi._n += 1
            retvals = func(*args, **kwargs)
            doi._n -= 1
            if doi._n == 0:
                # We have reached the point where the original `dois` argument
                # was passed. Create a new set such that subsequent calls don't
                # contaminate the bibliography.
                doi.dois = set()
            return retvals

        return func_with_doi



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

length_units = {'Gm': 1e9, 'Mm': 1e6, 'km': 1000.0, 'm': 1.0, 'mm': 1e-3, 'µm': 1e-6, 'um': 1e-6, 'nm': 1e-9,
                'Å': 1e-10, 'pm': 1e-12, 'fm': 1e-15}
voltage_units = {'GV': 1e9, 'MV': 1e6, 'kV': 1000.0, 'V': 1.0, 'mV': 1e-3, 'µV': 1e-6, 'nV': 1e-9, 'pV': 1e-12,
                 'fV': 1e-15}

units = dict(length=length_units, voltage=voltage_units)


def get_unit_conversion_factor(unit1_str, unit2_str):
    """
    Compute factor for conversion from unit1 to unit2.

    Parameters
    ----------
    unit1_str : str
        Name of source unit
    unit2_str : str
        Name of targe unit

    Returns
    -------
    fac : float
        Unit conversion factors. A quantity in unit1 is converted to unit2
        by multiplication with this factor.
    """
    if unit1_str is None:
        raise ValueError('Cannot convert from None unit')
    if unit2_str is None:
        raise ValueError('Cannot convert to None unit')
    if unit1_str == unit2_str:
        return 1
    unit1_kind = None
    unit2_kind = None
    unit_scales = None
    for key, values in units.items():
        if unit1_str in values:
            unit1_kind = key
            unit_scales = values
        if unit2_str in values:
            unit2_kind = key
            unit_scales = values
    if unit1_kind is None:
        raise ValueError(f"Unknown unit '{unit1_str}'.")
    if unit2_kind is None:
        raise ValueError(f"Unknown unit '{unit2_str}'.")
    if unit1_kind != unit2_kind:
        raise ValueError(f"Unit '{unit1_str}' is of kind {unit1_kind} while unit '{unit2_str}' is of kind {unit2_str}."
                         "I cannot convert between the two.")
    return unit_scales[unit1_str] / unit_scales[unit2_str]


def mangle_length_unit_utf8(unit):
    """
    Convert unit string to normalized UTF-8 unit string, e.g. converts 'um'
    to 'µm' and makes sure 'µ' is MICRO SIGN (00B5) and not GREEK SMALL LETTER
    MU (03BC).

    Parameters
    ----------
    unit : str
        Name of unit

    Returns
    -------
    output_unit : str
        Mangled name of unit
    """
    if isinstance(unit, str):
        unit = unit.strip()
    else:
        unit = unit.decode('utf-8').strip()
    if unit == '':
        return None
    elif unit == 'A':
        return 'Å'
    elif unit == 'μm' or unit == 'um' or unit == '~m':
        return 'µm'
    else:
        return unit


def mangle_length_unit_ascii(unit):
    """
    Convert unit string to ASCII representation, e.g. converts 'µm'
    to 'um'.

    Parameters
    ----------
    unit : str
        Name of unit

    Returns
    -------
    output_unit : str
        Mangled name of unit
    """
    unit = unit.strip()
    if unit == '':
        return None
    elif unit == 'Å':
        return 'A'
    elif unit == 'μm' or unit == 'µm' or unit == '~m':
        return 'um'
    else:
        return unit


def suggest_length_unit(scale, lower_in_meters, upper_in_meters):
    """
    Suggest a length unit for representing data in a certain range.
    E.g. data in the range from 1e-3 to 1e-2 m is best represented by um.

    Parameters
    ----------
    scale : str
        'linear': displaying data on a linear axis
        'log' displaying data on a log-space axis
    lower_in_meters : float
        Lower bound of range in meters
    upper_in_meters : float
        Upper bound of range in meters

    Returns
    -------
    unit : str
        Suggestion for the length unit
    """
    if scale == 'linear':
        v = max(abs(lower_in_meters), abs(upper_in_meters))
        m10 = 3 * int(np.floor(np.log10(v) / 3))
    elif scale == 'log':
        u10 = int(np.ceil(np.log10(upper_in_meters)))
        l10 = int(np.floor(np.log10(lower_in_meters)))
        m10 = 3 * int(np.ceil((l10 + u10) / 6) - 1)
    else:
        raise ValueError(f"Unknown scale parameter '{scale}'.")

    fac = 10 ** m10
    minfac = 1
    minunit = 'm'
    maxfac = 1
    maxunit = 'm'
    for key, value in length_units.items():
        if value < minfac:
            minfac = value
            minunit = key
        if value > maxfac:
            maxfac = value
            maxunit = key
        if value == fac:
            return key

    # We could not identify any unit from our list
    if fac < minfac:
        return minunit
    elif fac > maxfac:
        return maxunit
    else:
        raise RuntimeError(f'Cannot find unit for scale prefix {fac}.')


def suggest_length_unit_for_data(scale, data, unit):
    """
    Suggest a length unit for representing a data set.

    Parameters
    ----------
    scale : str
        'linear': displaying data on a linear axis
        'log' displaying data on a log-space axis
    data : array_like
        Data set that needs representing (e.g. in a color bar)
    unit : str
        Unit of the data set

    Returns
    -------
    unit : str
        Suggestion for the length unit
    """
    mn, mx = np.nanmin(data), np.nanmax(data)
    if np.isnan(mn) or np.isnan(mx):
        raise ValueError('Data does not contain values that are not NaN.')
    if not (np.isfinite(mn) and np.isfinite(mx)):
        # There are Inf or -Inf values in the data, we just return the original unit.
        return unit
    fac = get_unit_conversion_factor(unit, 'm')
    return suggest_length_unit(scale, fac * mn, fac * mx)



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# Reference information and implementations:
# https://www.osti.gov/biblio/1419732
# https://sourceforge.net/p/gwyddion/code/HEAD/tree/trunk/gwyddion/modules/file/keyence.c
# https://github.com/torkian/vk4-python-driver
#

import datetime
from zipfile import ZipFile

import numpy as np

from .binary import decode
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo
from ..Exceptions import CorruptFile, FileFormatMismatch, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography


class VKReader(ReaderBase):
    _format = 'vk'
    _name = 'Keyence VK'
    _description = '''
VK3, VK4, VK6 and VK7 file formats of the Keyence laser confocal microscope.
'''

    _MAGIC3 = b'VK3_'
    _MAGIC4 = b'VK4_'
    _MAGIC0 = b'\x00\x00\x00\x00'

    _MAGIC6 = b'VK6'
    _MAGIC7 = b'VK7'

    _offset_table_structure = [
        ('setting', 'I'),
        ('color_peak', 'I'),
        ('color_light', 'I'),
        ('light1', 'I'),
        ('light2', 'I'),
        ('light3', 'I'),
        ('height1', 'I'),
        ('height2', 'I'),
        ('height3', 'I'),
        ('color_peak_thumbnail', 'I'),
        ('color_thumbnail', 'I'),
        ('light_thumbnail', 'I'),
        ('height_thumbnail', 'I'),
        ('assemble', 'I'),
        ('line_measure', 'I'),
        ('line_thickness', 'I'),
        ('string_data', 'I'),
    ]

    _header_structure = [
        ('size', 'I'),
        ('year', 'I'),
        ('month', 'I'),
        ('day', 'I'),
        ('hour', 'I'),
        ('minute', 'I'),
        ('second', 'I'),
        ('diff_utc_by_minutes', 'i'),
        ('image_attributes', 'I'),
        ('user_interface_mode', 'I'),
        ('color_composite_mode', 'I'),
        ('num_layer', 'I'),
        ('run_mode', 'I'),
        ('peak_mode', 'I'),
        ('sharpening_level', 'I'),
        ('speed', 'I'),
        ('distance', 'I'),
        ('pitch', 'I'),
        ('optical_zoom', 'I'),
        ('num_line', 'I'),
        ('line0_pos', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('lens_mag', 'I'),
        ('pmt_gain_mode', 'I'),
        ('pmt_gain', 'I'),
        ('pmt_offset', 'I'),
        ('nd_filter', 'I'),
        ('_', 'I'),
        ('persist_count', 'I'),
        ('shutter_speed_mode', 'I'),
        ('shutter_speed', 'I'),
        ('white_balance_mode', 'I'),
        ('white_balance_red', 'I'),
        ('white_balance_blue', 'I'),
        ('camera_gain', 'I'),
        ('plane_compensation', 'I'),
        ('length_unit', 'I'),
        ('height_unit', 'I'),
        ('xy_decimal_place', 'I'),
        ('height_decimal_place', 'I'),
        ('x_length_per_pixel', 'I'),
        ('y_length_per_pixel', 'I'),
        ('height_scale_factor', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('light_filter_type', 'I'),
        ('_', 'I'),
        ('gamma_reverse', 'I'),
        ('gamma', 'I'),
        ('gamma_offset', 'I'),
        ('ccd_bw_offset', 'I'),
        ('numerical_aperture', 'I'),
        ('head_type', 'I'),
        ('pmt_gain2', 'I'),
        ('omit_color_image', 'I'),
        ('lens_id', 'I'),
        ('light_lut_mode', 'I'),
        ('light_lut_in0', 'I'),
        ('light_lut_out0', 'I'),
        ('light_lut_in1', 'I'),
        ('light_lut_out1', 'I'),
        ('light_lut_in2', 'I'),
        ('light_lut_out2', 'I'),
        ('light_lut_in3', 'I'),
        ('light_lut_out3', 'I'),
        ('light_lut_in4', 'I'),
        ('light_lut_out4', 'I'),
        ('upper_position', 'I'),
        ('lower_position', 'I'),
        ('light_effective_itemsize', 'I'),
        ('height_effective_itemsize', 'I'),
    ]

    _image_structure = [
        ('width', 'I'),
        ('height', 'I'),
        ('itemsize', 'I'),
        ('compression', 'I'),
        ('byte_size', 'I'),
        ('palette_range_min', 'I'),
        ('palette_range_max', 'I'),
    ]

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self.file_path = file_path
        self._file_version = None
        self.read_vk3467_header(file_path)

    def read_vk3467_header(self, file_path):
        with OpenFromAny(file_path, 'rb') as f:
            # Detect file version
            magic = f.read(4)

            # Check VK* file magic
            if magic.startswith(self._MAGIC3):  # 'VK3_'
                file_version = 3
            elif magic.startswith(self._MAGIC4):  # 'VK4_'
                file_version = 4
            elif magic.startswith(self._MAGIC6):  # 'VK6'
                file_version = 6
            elif magic.startswith(self._MAGIC7):  # 'VK7'
                file_version = 7
            else:
                raise FileFormatMismatch('File magic does not match. This is not a Keyence VK file.')

            if self._file_version is None:
                self._file_version = file_version

            if file_version in [3, 4]:
                f.read(4)  # skip dll version
                if f.read(len(self._MAGIC0)) != self._MAGIC0:  # All zeros
                    raise FileFormatMismatch('File magic does not match. I thought this was a Keyence VK3 or VK4 file, '
                                             'but it seems this is not the case.')
                self.read_vk34_header(f)
            else:
                # VK6/7 contains a .zip file that has VK4 file name 'Vk4File'.
                # ZipFile gracefully skips VK6/7 header information before the
                # zip actually starts.
                with ZipFile(f, 'r') as z:
                    self.read_vk3467_header(z.open('Vk4File'))

    def read_vk34_header(self, f):
        # Offset table
        self._offset_table = decode(f, self._offset_table_structure, '<')

        # File version 4 (and 6, 7, which is 4) have an additional entry here
        if self._file_version in [4, 6, 7]:
            f.read(4)

        # Measurement conditions
        self._header = decode(f, self._header_structure, '<')

        # Right now, we are assuming that there is only a single (height)
        # channel per VK4 file. Not sure if this is correct.

        # Height data
        f.seek(self._offset_table['height1'])
        self._data = decode(f, self._image_structure, '<')

        if self._data['byte_size'] != self._data['width'] * self._data['height'] * self._data['itemsize'] // 8:
            raise CorruptFile('Reported size does not match image dimensions.')

        if self._data['itemsize'] not in [8, 16, 32]:
            raise CorruptFile('Reported item size is {} bits, expected 8, 16 or 32 bits.'
                              .format(self._data['itemsize']))

        self._physical_sizes = ((self._data['width'] - 1) * self._header['x_length_per_pixel'],
                                (self._data['height'] - 1) * self._header['y_length_per_pixel'])
        self._unit = 'pm'

        self._info = {
            'acquisition_time':
                str(datetime.datetime(
                    self._header['year'], self._header['month'], self._header['day'],
                    self._header['hour'], self._header['minute'], self._header['second'],
                    tzinfo=datetime.timezone(datetime.timedelta(minutes=self._header['diff_utc_by_minutes']))
                )),
            'raw_metadata': self._header
        }

    def read_height_data(self, f):
        f.seek(self._offset_table['height1'])
        image_header = decode(f, self._image_structure, '<')
        f.seek(768, 1)  # Skip palette
        assert image_header['width'] == self._data['width']
        assert image_header['height'] == self._data['height']
        dtype = np.dtype('<u1') if self._data['itemsize'] == 8 \
            else np.dtype('<u2') if self._data['itemsize'] == 16 \
            else np.dtype('<u4')
        buffer = f.read(self._data['width'] * self._data['height'] * np.dtype(dtype).itemsize)
        return np.frombuffer(buffer, dtype=dtype).reshape((self._data['height'], self._data['width'])).T

    @property
    def channels(self):
        return [ChannelInfo(self,
                            0,  # channel index
                            name='Default',
                            dim=2,
                            nb_grid_pts=(self._data['width'], self._data['height']),
                            physical_sizes=self._physical_sizes,
                            uniform=True,
                            unit=self._unit)]

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI parallelization.')

        if channel_index is None:
            channel_index = self._default_channel_index

        if channel_index != self._default_channel_index:
            raise RuntimeError(f'There is only a single channel. Channel index must be {self._default_channel_index}.')

        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        with OpenFromAny(self.file_path, 'rb') as f:
            if self._file_version in [3, 4]:  # VK3 or VK4
                height_data = self.read_height_data(f)
            else:  # VK6 or VK7
                # ZipFile gracefully skips VK6 header information before the
                # zip actually starts.
                with ZipFile(f) as z:
                    height_data = self.read_height_data(z.open('Vk4File'))

        _info = self._info.copy()
        _info.update(info)

        topo = Topography(height_data,
                          self._physical_sizes,
                          unit=self._unit,
                          periodic=False if periodic is None else periodic,
                          info=_info)
        return topo.scale(float(self._header['height_scale_factor']))



#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
import openpyxl
import re

from datetime import datetime

from ..Exceptions import MetadataAlreadyFixedByFile

from ..NonuniformLineScan import NonuniformLineScan
from ..UniformLineScanAndTopography import UniformLineScan
from ..Support.UnitConversion import get_unit_conversion_factor
from .Reader import ReaderBase, ChannelInfo

R_metric_regex = re.compile(r'(?P<key>[a-zA-Z]+)\s+(?P<value>[+-]?(?:[0-9]*[.])?[0-9]+)\s+(?P<unit>[^\s]+)')
cut_off_regex = re.compile(r'(?P<value>[+-]?(?:[0-9]*[.])?[0-9]+)\s*(?P<unit>[^\s]+)')


class MitutoyoReader(ReaderBase):
    """
    Mitutoyo SurfTest surface roughness testers produce specifically formatted
    Excel spreadsheets.
    """

    _format = 'mitutoyo'
    _name = 'Mitutoyo SurfTest Excel spreadsheet (xlsx)'
    _description = '''
Load topography information stored as Excel spreadsheet by Mitutoyo SurfTest
surface roughness testers.
    '''

    def __init__(self, fobj):
        """
        Open Excel spreadsheet produced by Mitutoyo SurfTest surface roughness
        testers.

        The reader expects a line scan by positions and heights in columns
        5 and 6 (E, F) and tries to extract standard roughness metrics from
        column 1 (A) on the sheet 'DATA'. The reader expects the
        acquisition date in column 5 (E) row 2 on sheet 'Certificate'.

        Parameters
        ----------
        fobj : filename or file object
             File to read.
        """

        try:
            wb = openpyxl.load_workbook(fobj)
            _data = wb['DATA']
            _metadata = wb['Certificate']
            # How should a reader properly fulfill test_no_resource_warning_on_failure?
        except openpyxl.utils.exceptions.InvalidFileException as exc:
            # reraise openpyxl exceptions as IOError
            raise IOError(exc)
        else:
            _x_cells, _h_cells = _data['E:F']
            _x = np.array([cell.value for cell in _x_cells if cell.value is not None])
            # We store the profile (and metadata) in the object since there
            # can only be a single data channel.
            self._h = np.array([cell.value for cell in _h_cells if cell.value is not None])

            # check if positions are distributed uniformly
            _diff = np.diff(_x)
            if np.all(np.isclose(_diff, _diff[0])):
                self._uniform = True
            else:
                self._uniform = False
                # height values are assigned to the "end" of each "pixel" in the
                # xlsx format, starting at non-zero x. Here we used to shift
                # positions by half a distance each to have non-uniform profiles
                # element-centered positions:
                # _x = _x - np.insert(_diff, 0, _x[0]) * 0.5
                # However, UniformLinescan always starts at x0 = 0
                # To conform with this convention, here we now simply remove the
                # first grid point's absolute position.
                _x -= _x[0]

            # try extracting simple roughness metrics from first column

            _roughness_metrics_cells = _data['A']
            _roughness_metrics_list = []
            for cell in _roughness_metrics_cells:
                # iterate until first empty row
                if cell.value is None:
                    break

                _roughness_metrics_record = {
                    key: float(value) if key == 'value' else value
                    for key, value in R_metric_regex.match(cell.value).groupdict().items()
                }

                _roughness_metrics_list.append(_roughness_metrics_record)

            # try to infer heights unit from roughness metrics
            _h_unit = _roughness_metrics_list[0]['unit']

            # get creation date
            _date_string = _metadata['E2'].value

            # remove all whitespace from date string
            _date_string = re.sub(r"\s+", "", _date_string, flags=re.UNICODE)

            # try to infer x unit from cut off
            _cut_off_string = _metadata['E48'].value
            _cut_off_dict = cut_off_regex.match(_cut_off_string).groupdict()
            _x_unit = _cut_off_dict['unit']

            # convert x unit to h unit
            self._x = _x*get_unit_conversion_factor(_x_unit, _h_unit)
            self._unit = _h_unit

            # with n data points spaced by distance dx,
            #   internal format of positions in uniform line scan is [0, dx, 2*dx, ... (n-1)*dx]
            #   Mitutoyo format of positions is [dx, 2*dx, ..., n*dx]
            # _physical_sizes must capture n*dx to result in valid zero-initiated uniform distribution
            self._physical_sizes = np.max(self._x)

            self._info = {
                'roughness_metrics': _roughness_metrics_list,
                'cut_off': _cut_off_dict,
                'acquisition_time': str(datetime.strptime(_date_string, '%d-%b-%Y'))
            }
            self._channels = [ChannelInfo(self,
                                          0,  # channel index
                                          # Since the is only a single channel and the file has no channel information,
                                          # the name is 'Default'
                                          name='Default',
                                          dim=1,
                                          uniform=self._uniform,
                                          nb_grid_pts=len(self._h),
                                          unit=self._unit,
                                          physical_sizes=self._physical_sizes,
                                          info=self._info)
                              ]

    # Return list of channels (here only a single one)
    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):

        # Check if only a subdomain should be loaded
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')

        # Check that channel_index is valid
        if channel_index is not None and channel_index != 0:
            raise ValueError('`channel_index` must be None or 0.')

        # If no channel index is given, we use the default index
        if channel_index is None:
            channel_index = self._default_channel_index

        # Units are specified in the XLSX file and cannot be overridden
        if unit is not None:
            if self._unit is not None:
                raise MetadataAlreadyFixedByFile('unit')
        else:
            unit = self._unit
        # Augment info dictionary with user-specified data
        _info = self._info.copy()
        _info.update(info)

        # Get channel information (there is only one)
        channel = self._channels[channel_index]

        # Make sure `physical_sizes` is present, either fixed by the file
        # or given by the user. Also make sure that if the user specifies it,
        # it cannot be set in the file. (We cannot override metadata from
        # files.)
        physical_sizes = self._check_physical_sizes(physical_sizes,
                                                    channel.physical_sizes)

        # may a reader return either NonuniformLineScan or UniformLineScan?
        if self._uniform:
            # Return a uniform line scan if data is equally spaced
            topography = UniformLineScan(
                self._h, physical_sizes,
                periodic=False if periodic is None else periodic,
                unit=unit,
                info=_info)
        else:
            if periodic is not None and periodic:
                raise ValueError('Mitutoyo reader found nonuniform data, and the user specified that it is periodic. '
                                 'Nonuniform line scans cannot be periodic.')

            # Return a nonuniform line scan otherwise
            topography = NonuniformLineScan(
                self._x,
                self._h,
                unit=unit,
                info=_info)

        # Check if there is a user-specified height scale factor
        if height_scale_factor is not None:
            topography = topography.scale(height_scale_factor)

        return topography

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__



#
# Copyright 2018, 2020 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Helper functions to compute trends of surfaces
"""

import numpy as np


def polyfit(x, h, deg):
    r"""
    Compute the detrending plane that, if subtracted, minimizes the rms height
    of the surface. The detrending plane is parameterized as a polynomial:

    .. math::

        p(x) = \sum_{k=0}^n a_k x^k

    The values of :math:`a_k` are returned by this function.

    The rms height of the surface is given by (see `rms_height`)

    .. math::

        h_\text{rms}^2 = \frac{1}{3L} \sum_{i=0}^{N-2} \left( h_i^2 + h_{i+1}^2 + h_i h_{i+1} \right) \Delta x_i

    where :math:`N` is the total number of data points. Hence we need to solve the following minimization problem:

    .. math::

        \min_{\{a_k\}} \left\{ \frac{1}{3L} \sum_{i=0}^{N-2} \left[ (h_i - p(x_i))^2 + (h_{i+1} - p(x_{i+1}))^2 + (h_i - p(x_i))(h_{i+1} - p(x_{i+1})) \right] \Delta x_i \right\}

    This gives the system of linear equations (one for each :math:`k`)

    .. math::

        \sum_{i=0}^{N-2} \left( \left[ 2h_i + h_{i+1} \right] x_i^k + \left[ 2h_{i+1} + h_i \right] x_{i+1}^k \right) \Delta x_i = \sum_{l=0}^n a_l \sum_{i=0}^{N-2} \left( 2x_i^{k+l} + 2x_{i+1}^{k+l} + x_i^k x_{i+1}^l + x_{i+1}^k x_i^l \right) \Delta x_i

    Parameters
    ----------
    x : array_like
        Array containing positions. This function assumes that this array is
        sorted in ascending order.
    h : array_like
        Array containing heights.
    deg : int
        Degree of polynomial :math:`n`.

    Returns
    -------
    a : array
        Array with coefficients :math:`a_k`.
    """  # noqa: E501
    dx = np.diff(x)
    k = np.arange(deg + 1).reshape(-1, 1)
    b = np.sum(((2 * h[:-1] + h[1:]) * x[:-1] ** k +
                (2 * h[1:] + h[:-1]) * x[1:] ** k) * dx,
               axis=1)
    L = k.reshape(1, -1, 1)
    k = k.reshape(-1, 1, 1)
    A = np.sum((2 * x[:-1] ** (k + L) + 2 * x[1:] ** (k + L) +
                x[:-1] ** k * x[1:] ** L + x[1:] ** k * x[:-1] ** L) * dx,
               axis=2)
    return np.linalg.solve(A, b)



#
# Copyright 2017-2021 Lars Pastewka
#           2018-2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
defines all surface types used in SurfaceTopography
"""

from .Container import SurfaceContainer, read_container, read_published_container  # noqa: F401
from .IO import open_topography, read_topography  # noqa: F401
from .NonuniformLineScan import NonuniformLineScan  # noqa: F401
from .Special import make_sphere, PlasticTopography  # noqa: F401
from .UniformLineScanAndTopography import Topography, UniformLineScan  # noqa: F401

# These imports are required to register the analysis functions!
import SurfaceTopography.Generic.Curvature  # noqa: F401
import SurfaceTopography.Generic.ReliabilityCutoff  # noqa: F401
import SurfaceTopography.Generic.ScaleDependentStatistics  # noqa: F401
import SurfaceTopography.Generic.ScanningProbe  # noqa: F401
import SurfaceTopography.Generic.Slope  # noqa: F401
import SurfaceTopography.Nonuniform.common  # noqa: F401
import SurfaceTopography.Nonuniform.Autocorrelation  # noqa: F401
import SurfaceTopography.Nonuniform.Converters  # noqa: F401
import SurfaceTopography.Nonuniform.Derivative  # noqa: F401
import SurfaceTopography.Nonuniform.Interpolation  # noqa: F401
import SurfaceTopography.Nonuniform.PowerSpectrum  # noqa: F401
import SurfaceTopography.Nonuniform.ScalarParameters  # noqa: F401
import SurfaceTopography.Nonuniform.VariableBandwidth  # noqa: F401
import SurfaceTopography.Uniform.common  # noqa: F401
import SurfaceTopography.Uniform.Autocorrelation  # noqa: F401
import SurfaceTopography.Uniform.Converters  # noqa: F401
import SurfaceTopography.Uniform.Derivative  # noqa: F401
import SurfaceTopography.Uniform.Filtering  # noqa: F401
import SurfaceTopography.Uniform.Imputation  # noqa: F401
import SurfaceTopography.Uniform.Interpolation  # noqa: F401
import SurfaceTopography.Uniform.PowerSpectrum  # noqa: F401
import SurfaceTopography.Uniform.ScalarParameters  # noqa: F401
import SurfaceTopography.Uniform.VariableBandwidth  # noqa: F401
import SurfaceTopography.Support.Bibliography  # noqa: F401

# Add contact.engineering paper to bibliography
SurfaceTopography.Support.Bibliography._default_dois = set(['10.1088/2051-672X/ac860a'])

try:
    from importlib.metadata import version

    __version__ = version(__name__)
except ImportError:
    from pkg_resources import get_distribution

    __version__ = get_distribution(__name__).version



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from struct import unpack

import numpy as np

from ..Exceptions import MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography

from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo


def mask_undefined(data, maxval=1e32):
    """
    If data contains undefined points, then return a masked array with all
    undefined points masked.

    The following heuristics is applied to identify undefined points:
    - Remove points that are +/-inf or nan
    - Remove points that are >+maxval or <-maxval
    """
    # First, we mask all points that are infinite or nan
    mask = np.logical_and(np.isfinite(data), np.abs(data) < maxval)
    if mask.sum() < len(mask):
        return np.ma.masked_array(data, mask=np.logical_not(mask))
    else:
        return data


class OPDReader(ReaderBase):
    _format = 'opd'
    _name = 'Wyko OPD'
    _description = '''
Files generated by the Vision software of the Bruker Wyko white-light
interferometer.
'''

    # Reads in the positions of all the data and metadata
    def __init__(self, fobj):
        self._fobj = fobj

        channel_index = 0
        self._channels = []

        with OpenFromAny(fobj, 'rb') as f:
            BLOCK_SIZE = 24

            def read_block(f):
                blkname = f.read(16).split(b'\0', 1)[0].decode('latin-1')
                blktype, blklen, blkattr = unpack('<hlH', f.read(8))
                return blkname, blktype, blklen, blkattr

            # Skip header
            f.read(2)

            # Read directory block
            dirname, dirtype, dirlen, dirattr = read_block(f)
            if dirname != 'Directory':
                raise IOError("Error reading directory block. "
                              "Header is '{}', expected 'Directory'".format(dirname))
            num_blocks = dirlen // BLOCK_SIZE
            if num_blocks * BLOCK_SIZE != dirlen:
                raise IOError('Directory length is not a multiple of the block physical_sizes.')

            blocks = []
            for i in range(num_blocks - 1):
                blocks += [read_block(f)]

            pixel_size = 1.0
            aspect = 1.0
            mult = 1.0
            for n, t, L, a in blocks:
                if L <= 0:
                    continue
                if n == 'RAW DATA' or n == 'RAW_DATA' or n == 'OPD' or n == 'Raw':
                    nx, ny, elsize = unpack('<HHH', f.read(6))
                    if elsize == 1:
                        dtype = np.dtype('c')
                    elif elsize == 2:
                        dtype = np.dtype('<i2')
                    elif elsize == 4:
                        dtype = np.dtype('f4')
                    else:
                        raise IOError("Don't know how to handle element of size {}."
                                      .format(elsize))

                    self._channels += [ChannelInfo(self, channel_index,
                                                   name=n,
                                                   dim=2,
                                                   nb_grid_pts=(nx, ny),
                                                   uniform=True,
                                                   unit='mm',
                                                   info=dict(file_offset=f.tell(), dtype=dtype))]
                    channel_index += 1

                    # Skip over data buffer
                    f.seek(nx * ny * dtype.itemsize, 1)
                elif n == 'Wavelength':
                    wavelength, = unpack('<f', f.read(4))
                elif n == 'Mult':
                    mult, = unpack('<H', f.read(2))
                elif n == 'Aspect':
                    aspect, = unpack('<f', f.read(4))
                elif n == 'Pixel_size':
                    pixel_size, = unpack('<f', f.read(4))
                else:
                    f.read(L)

        # Loop over all channels and adjust physical sizes and height scale factor
        for channel in self._channels:
            nx, ny = channel.nb_grid_pts
            channel.physical_sizes = (nx * pixel_size, ny * pixel_size * aspect)
            channel.height_scale_factor = wavelength / mult * 1e-6

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={}, periodic=False,
                   subdomain_locations=None, nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        channel = self._channels[channel_index]

        nb_pixels = np.prod(channel.nb_grid_pts)
        dtype = channel.info['dtype']
        with OpenFromAny(self._fobj, 'rb') as f:
            f.seek(channel.info['file_offset'])
            rawdata = f.read(nb_pixels * dtype.itemsize)
        data = np.frombuffer(rawdata, count=nb_pixels, dtype=dtype)
        data = mask_undefined(data)
        data.shape = channel.nb_grid_pts

        # Height are in nm, width in mm
        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')
        topography = Topography(np.fliplr(data), channel.physical_sizes, unit='mm', info=info, periodic=periodic)
        return topography.scale(channel.height_scale_factor)

    @property
    def channels(self):
        return self._channels

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__



#
# Copyright 2018-2021 Lars Pastewka
#           2019 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Variable bandwidth analysis for nonuniform topographies
"""

import numpy as np

from ..Exceptions import NoReliableDataError
from ..HeightContainer import NonuniformLineScanInterface
from ..NonuniformLineScan import NonuniformLineScan
from ..Support import build_tuple


def checkerboard_detrend_profile(self, subdivisions, tol=1e-6):
    """
    Perform tilt correction (and substract mean value) in each individual
    rectangle of a checkerboard decomposition of the surface. This is
    identical to subdividing the surface into individual, nonoverlapping
    rectangles and performing individual tilt corrections on them.

    The main application of this function is to carry out a variable
    bandwidth analysis of the surface.

    Parameters
    ----------
    self : :obj:`NonuniformLineScan`
        Container storing the uniform topography map
    subdivisions : int
        Number of subdivisions.
    tol : float
        Tolerance for searching for existing data points at domain boundaries.
        (Default: 1e-6)

    Returns
    -------
    subdivided_line_scans : list of :obj:`NonuniformLineScan`
        List with new, subdivided and detrended line scans.
    """
    if subdivisions == 1:
        return [self.detrend()]

    x, y = self.positions_and_heights()

    subdivided_line_scans = []
    for i in range(subdivisions):
        # Subdivide interval
        sub_xleft = x[0] + i * (x[-1] - x[0]) / subdivisions
        sub_xright = x[0] + (i + 1) * (x[-1] - x[0]) / subdivisions

        # Search for the data point closes to sub_xleft and sub_xright
        sub_ileft = x.searchsorted(sub_xleft)
        sub_iright = x.searchsorted(sub_xright, side='right')

        sub_x = x[sub_ileft:sub_iright]
        sub_y = y[sub_ileft:sub_iright]

        # Put additional data points on the left and right boundaries, if there
        # is none already in the data set at exactly those points

        if sub_ileft != 0 and sub_xleft < x[sub_ileft] - tol:
            # Linear interpolation to boundary point
            sub_yleft = y[sub_ileft - 1] + (sub_xleft - x[sub_ileft - 1]) / (
                        x[sub_ileft] - x[sub_ileft - 1]) * (
                                y[sub_ileft] - y[sub_ileft - 1])
            # Add additional point to data
            sub_x = np.append([sub_xleft], sub_x)
            sub_y = np.append([sub_yleft], sub_y)

        if sub_iright != len(x) and sub_xright > x[sub_iright - 1] + tol:
            # Linear interpolation to boundary point
            sub_yright = y[sub_iright - 1] + (
                        sub_xright - x[sub_iright - 1]) / (
                                     x[sub_iright] - x[sub_iright - 1]) * (
                                 y[sub_iright] - y[sub_iright - 1])
            # Add additional point to data
            sub_x = np.append(sub_x, [sub_xright])
            sub_y = np.append(sub_y, [sub_yright])

        subdivided_line_scans += [
            NonuniformLineScan(sub_x, sub_y, info=self.info).detrend()]

    return subdivided_line_scans


def variable_bandwidth_from_profile(self, quantities='bh', reliable=True, resampling_method=None,
                                    nb_grid_pts_cutoff=4):
    """
    Perform a variable bandwidth analysis by computing the mean
    root-mean-square height within increasingly finer subdivisions of the
    line scan.

    Parameters
    ----------
    self : obj:`NonuniformLineScan`
        Container storing the uniform topography map
    quantities : str, optional
        Specification of return tuple, where each string character
        stand for specific quantity. Possible quantities are
            - 'm': Magnification (Unit: dimensionless)
            - 'b': Bandwidth (Unit: length)
            - 'h': RMS height (Unit: length)
        For example, 'mbh' return a tuple with the three entries
        magnification, bandwidth, rms height.
        (Default: 'bh')
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    resampling_method : str, optional
        Can only be None for the variable bandwidth analysis.
        (Default: None)
    nb_grid_pts_cutoff : int
        Minimum number of data points to allow for subdivision. The analysis
        will automatically analyze subdivision down to this nb_grid_pts.

    Returns
    -------
    magnifications : array
        Array containing the magnifications.
    bandwidths : array
        Array containing the bandwidths, here the physical_sizes of the
        subdivided topography.
    rms_heights : array
        Array containing the rms height corresponding to the respective
        magnification.
    """
    if resampling_method is not None:
        raise ValueError('`variable_bandwidth_from_profile` does not support resampling.')

    magnification = 1
    min_nb_grid_pts, = self.nb_grid_pts
    magnifications = []
    bandwidths = []
    rms_heights = []
    while min_nb_grid_pts >= nb_grid_pts_cutoff:
        subdivided_line_scans = self.checkerboard_detrend_profile(magnification)
        min_nb_grid_pts = min(
            [line.nb_grid_pts[0] for line in subdivided_line_scans])
        magnifications += [magnification]
        bandwidths += [subdivided_line_scans[0].physical_sizes[0]]
        rms_heights += [
            np.mean([line.rms_height_from_profile() for line in subdivided_line_scans])]
        magnification *= 2

    magnifications = np.array(magnifications)
    bandwidths = np.array(bandwidths)
    rms_heights = np.array(rms_heights)

    if reliable:
        short_cutoff = self.short_reliability_cutoff()
        if short_cutoff:
            m = bandwidths > short_cutoff
            if m.sum() == 0:
                raise NoReliableDataError('Dataset contains no reliable data.')
            magnifications = magnifications[m]
            bandwidths = bandwidths[m]
            rms_heights = rms_heights[m]

    return build_tuple(quantities, m=np.array(magnifications), b=np.array(bandwidths), h=np.array(rms_heights))


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('checkerboard_detrend_profile', checkerboard_detrend_profile)
NonuniformLineScanInterface.register_function('variable_bandwidth_from_profile', variable_bandwidth_from_profile)



#
# Copyright 2018-2020 Lars Pastewka
#           2019 Antoine Sanner
#           2019 Michael Röttger
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Compute derivatives of nonuniform line scans."""

import numpy as np

from ..Exceptions import ReentrantDataError
from ..HeightContainer import NonuniformLineScanInterface
from ..Support import toiter, fromiter


def derivative(self, n, scale_factor=None, distance=None, interpolation='linear', progress_callback=None):
    r"""
    Compute derivative of nonuniform line-scan. Function assumes nonperiodic
    topographies.

    First derivative: Central differences.

    Second derivative: Expand :math:`h(x+\Delta x_+)` and :math:`(x-\Delta x_-)` up to second order in the grid
    spacing :math:`\Delta x_+` and :math:`\Delta x_+`. Then
    :math:`\Delta x_- f(x+\Delta x_+) + \Delta x_+ f(x+\Delta x_-)` yields:

    .. math::

         \frac{d^2h}{dx^2} \approx 2 \frac{\Delta x_-\left[f(x+\Delta x_+)-f(x)\right] + \Delta x_+\left[f(x+\Delta x_-)-f(x)\right]}{\Delta x_+\Delta x_-(\Delta x_++\Delta x_-)}

    Parameters
    ----------
    self : :class:`SurfaceTopography.NonuniformLineScan`
        Object containing height information.
    n : int
        Number of times the derivative is taken.
    scale_factor : int or list of ints or list of tuples of ints, optional
        Scale factors are not supported by nonuniform line scans.
        (Default: None)
    distance : float or list of floats, optional
        Explicit distance scale for computation of the derivative. Note that
        the distance specifies the overall length of the stencil of lowest
        truncation order, not the effective grid spacing used by this stencil.
        The scale factor is then given by distance / (n * px) where n is the
        order of the derivative and px the grid spacing.
        (Default: None)
    interpolation : str, optional
        Only 'linear' interpolation is supported by nonuniform line scans.
        (Default: 'linear')
    progress_callback : func, optional
        Function taking iteration and the total number of iterations as
        arguments for progress reporting. (Default: None)

    Returns
    -------
    derivative : np.ndarray
        Array with derivative values. Length of array is reduced by :math:`n` with
        respect to the input array for the :math:`n`-th derivative.
    """  # noqa: E501
    if scale_factor is not None:
        raise ValueError('Scale factors are not supported by nonuniform line scans.')
    if interpolation != 'linear':
        raise ValueError('Line scans only support linear interpolation.')
    if self.is_reentrant:
        raise ReentrantDataError('This topography is reentrant (i.e. it contains overhangs). Derivatives cannot be '
                                 'computed for reentrant topographies.')

    if distance is not None:
        # If an explicit distance scale is given, we interpolate onto a regular grid and pass the derivative
        # calculation on to the uniform line scan derivative function.
        lower, upper = self.bandwidth()
        derivatives = []
        distances = toiter(distance)
        for i, d in enumerate(distances):
            if progress_callback is not None:
                progress_callback(i, len(distances))
            scale_factor = int(np.ceil(d / lower))
            stencil_size = d / scale_factor
            derivatives += [self.to_uniform(pixel_size=stencil_size / n).derivative(n=n, scale_factor=scale_factor,
                                                                                    interpolation='disable')]
        if progress_callback is not None:
            progress_callback(len(distances), len(distances))

        return fromiter(derivatives, distance)

    # If no explicit distance scale is given, we simply compute the derivatives from finite differences expressions on
    # nonuniform grids.
    x, h = self.positions_and_heights()
    if n == 1:
        return np.diff(h) / np.diff(x)
    elif n == 2:
        dxp = x[2:] - x[1:-1]
        dxm = x[1:-1] - x[:-2]

        return 2 * (dxm * (h[2:] - h[1:-1]) + dxp * (h[0:-2] - h[1:-1])) / (
                dxp * dxm * (dxp + dxm))
    else:
        raise RuntimeError('Currently only first and second derivatives are '
                           'supported for nonuniform topographies.')


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('derivative', derivative)



#
# Copyright 2019-2021 Lars Pastewka
#           2019, 2021 Michael Röttger
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Support for nonuniform topogography descriptions
"""

import numpy as np

from .HeightContainer import AbstractTopography, DecoratedTopography, \
    NonuniformLineScanInterface
from .Nonuniform.Detrending import polyfit
from .Support.UnitConversion import get_unit_conversion_factor


class NonuniformLineScan(AbstractTopography, NonuniformLineScanInterface):
    """
    Nonuniform topography with point list consisting of static numpy arrays.
    """

    def __init__(self, x, y, unit=None, info={}):
        """
        Constructor.

        Arguments
        ---------
        x : array_like
            x-positions of the data points that sample the line scan.
        y : array_like
            y-positions of the data points that sample the line scan.
        unit : str, optional
            The length unit.
        info : dict, optional
            The info dictionary containing auxiliary data.
        """
        super().__init__(unit=unit, info=info)
        if np.ma.is_masked(x) or np.ma.is_masked(y):
            raise ValueError('Storage arrays for nonuniform line scans cannot be masked.')
        self._x = np.asarray(x)
        self._h = np.asarray(y)

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._x, self._h
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._x, self._h = state
        super().__setstate__(superstate)

    # Implement abstract methods of AbstractHeightContainer

    @property
    def dim(self):
        return 1

    @property
    def physical_sizes(self):
        """Returns distance between maximum and minimum x-value."""
        return self._x[-1] - self._x[0],

    @property
    def is_periodic(self):
        """Return whether the topography is periodically repeated at the
        boundaries."""
        return False

    # Implement uniform line scan interface

    @property
    def nb_grid_pts(self):
        return len(self._x),

    @property
    def x_range(self):
        return self._x[0], self._x[-1]

    def positions(self):
        return self._x

    def heights(self):
        return self._h

    def squeeze(self):
        return self


class DecoratedNonuniformTopography(DecoratedTopography,
                                    NonuniformLineScanInterface):
    @property
    def is_periodic(self):
        return self.parent_topography.is_periodic

    @property
    def dim(self):
        return self.parent_topography.dim

    @property
    def nb_grid_pts(self):
        return self.parent_topography.nb_grid_pts

    @property
    def unit(self):
        if self._unit is None:
            return self.parent_topography.unit
        else:
            return self._unit

    @property
    def info(self):
        info = self.parent_topography.info
        info.update(self._info)
        if self.unit is not None:
            info.update(dict(unit=self.unit))
        return info

    @property
    def physical_sizes(self):
        return self.parent_topography.physical_sizes

    @property
    def x_range(self):
        return self.parent_topography.x_range

    def positions(self):
        return self.parent_topography.positions()

    def squeeze(self):
        return NonuniformLineScan(self.positions(), self.heights(),
                                  info=self.info)


class ScaledNonuniformTopography(DecoratedNonuniformTopography):
    """Scale heights, positions, or both."""

    def __init__(self, topography, unit, info={}):
        """
        This topography wraps a parent topography and rescales x and z
        coordinates according to certain rules.

        Arguments
        ---------
        topography : :obj:`NonuniformTopographyInterface`
            Parent topography
        unit : str
            Target unit.
        info : dict, optional
            Updated entries to the info dictionary. (Default: {})
        """
        super().__init__(topography, unit=unit, info=info)

    @property
    def height_scale_factor(self):
        return get_unit_conversion_factor(self.parent_topography.unit, self.unit)

    @property
    def position_scale_factor(self):
        return get_unit_conversion_factor(self.parent_topography.unit, self.unit)

    @property
    def physical_sizes(self):
        """Compute rescaled physical sizes."""
        return self.position_scale_factor * super().physical_sizes[0],

    def positions(self):
        """Compute the rescaled positions."""
        return self.position_scale_factor * super().positions()

    def heights(self):
        """ Computes the rescaled profile.
        """
        return self.height_scale_factor * self.parent_topography.heights()


class StaticallyScaledNonuniformTopography(ScaledNonuniformTopography):
    """Scale heights, positions, or both."""

    def __init__(self, topography, height_scale_factor, position_scale_factor=1, unit=None, info={}):
        """
        This topography wraps a parent topography and rescales x and z
        coordinates according to certain rules.

        Arguments
        ---------
        topography : :obj:`NonuniformTopographyInterface`
            Parent topography
        height_scale_factor : float
            Factor to scale heights with.
        position_scale_factor : float, optional
            Factor to scale lateral positions (`physical_sizes`, etc.) with.
            (Default: 1)
        unit : str, optional
            Target unit. This is simply used to update the metadata, not for
            determining scale factors. (Default: None)
        info : dict, optional
            Updated entries to the info dictionary. (Default: {})
        """
        super().__init__(topography, unit=unit, info=info)
        self._height_scale_factor = None if height_scale_factor is None else float(height_scale_factor)
        self._position_scale_factor = None if position_scale_factor is None else float(position_scale_factor)

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._height_scale_factor, self._position_scale_factor, self._unit
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._height_scale_factor, self._position_scale_factor, self._unit = state
        super().__setstate__(superstate)

    @property
    def height_scale_factor(self):
        return self._height_scale_factor

    @property
    def position_scale_factor(self):
        return self._position_scale_factor


class DetrendedNonuniformTopography(DecoratedNonuniformTopography):
    """
    Remove trends from a topography. This is achieved by fitting polynomials
    to the topography data to extract trend lines. The resulting topography
    is then detrended by substracting these trend lines.
    """

    def __init__(self, topography, detrend_mode='height', info={}):
        """
        Parameters
        ----------
        topography : SurfaceTopography
            SurfaceTopography to be detrended.
        detrend_mode : str
            'center': center the topography, no trend correction.
            'height': adjust slope such that rms height is minimized.
            'slope': adjust slope such that rms slope is minimized.
            'curvature': adjust slope and curvature such that rms height is
            minimized.
            (Default: 'height')
        """
        super().__init__(topography, info=info)
        self._detrend_mode = detrend_mode
        self._detrend()

    def _detrend(self):
        if self._detrend_mode == 'center':
            x, y = self.parent_topography.positions_and_heights()
            self._coeffs = polyfit(x, y, 0)
        elif self._detrend_mode == 'height':
            x, y = self.parent_topography.positions_and_heights()
            self._coeffs = polyfit(x, y, 1)
        elif self._detrend_mode == 'slope':
            sl = self.parent_topography.derivative(1).mean()
            self._coeffs = [self.parent_topography.mean(), sl]
        elif self._detrend_mode == 'curvature':
            x, y = self.parent_topography.positions_and_heights()
            self._coeffs = polyfit(x, y, 2)
        else:
            raise ValueError("Unsupported detrend mode '{}' for line scans.".format(self._detrend_mode))

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._detrend_mode, self._coeffs
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._detrend_mode, self._coeffs = state
        super().__setstate__(superstate)

    @property
    def coeffs(self, ):
        return self._coeffs

    @property
    def detrend_mode(self, ):
        return self._detrend_mode

    @detrend_mode.setter
    def detrend_mode(self, detrend_mode):
        self._detrend_mode = detrend_mode
        self._detrend()

    @property
    def is_periodic(self):
        """
        SurfaceTopography stays periodic only after detrend mode "center".
        Otherwise the detrended SurfaceTopography is non-periodic.
        """
        if self.detrend_mode == "center":
            return self.parent_topography.is_periodic
        else:
            return False

    @property
    def x_range(self):
        return self.parent_topography.x_range

    def positions(self):
        return self.parent_topography.positions()

    def heights(self):
        """ Computes the combined profile.
        """
        if len(self._coeffs) == 1:
            a0, = self._coeffs
            return self.parent_topography.heights() - a0
        x = self.positions()
        if len(self._coeffs) == 2:
            a0, a1 = self._coeffs
            return self.parent_topography.heights() - a0 - a1 * x
        elif len(self._coeffs) == 3:
            a0, a1, a2 = self._coeffs
            return self.parent_topography.heights() - a0 - a1 * x - a2 * x * x
        else:
            raise RuntimeError('Unknown physical_sizes of coefficients '
                               'tuple.')

    def stringify_plane(self, fmt=lambda x: str(x)):
        str_coeffs = [fmt(x) for x in self._coeffs]
        if len(self._coeffs) == 1:
            h0, = str_coeffs
            return h0
        elif len(self._coeffs) == 2:
            return '{0} + {1} x'.format(*str_coeffs)
        elif len(self._coeffs) == 3:
            return '{0} + {1} x + {2} x^2'.format(*str_coeffs)
        else:
            raise RuntimeError('Unknown physical_sizes of coefficients '
                               'tuple.')

    @property
    def curvatures(self):
        if len(self._coeffs) == 3:
            return 2 * self._coeffs[2],
        else:
            return 0,


# Register analysis functions from this module
NonuniformLineScanInterface.register_function(
    'mean', lambda this: np.trapz(this.heights(), this.positions()) / this.physical_sizes[0])
NonuniformLineScanInterface.register_function('min', lambda this: this.heights().min())
NonuniformLineScanInterface.register_function('max', lambda this: this.heights().max())

# Register pipeline functions from this module
NonuniformLineScanInterface.register_function('to_unit', ScaledNonuniformTopography)
NonuniformLineScanInterface.register_function('scale', StaticallyScaledNonuniformTopography)
NonuniformLineScanInterface.register_function('detrend', DetrendedNonuniformTopography)



#
# Copyright 2021 Lars Pastewka
#           2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Scale-dependent slope
"""

import numpy as np

from ..HeightContainer import NonuniformLineScanInterface, UniformTopographyInterface


def scale_dependent_slope_from_profile(topography, **kwargs):
    r"""
    Compute the one-dimensional scale-dependent slope.

    The scale-dependent slope is given by

       .. math::
         :nowrap:

         \begin{equation}
         h_\text{rms}^\prime(\lambda) = \left[2A(\lambda)\right]^{1/2}/\lambda
         \end{equation}

    where :math:`A(\lambda)` is the autocorrelation function.

    Parameters
    ----------
    topography : :class:`SurfaceTopography.Topography` or :class:`SurfaceTopography.UniformLineScan`
        Container storing the uniform topography map
    **kwargs : dict
        Additional keyword parameters are passed on to
        `autocorrelation_from_profile`

    Returns
    -------
    r : array
        Distances. (Units: length)
    slope : array
        Slope. (Units: dimensionless)
    """  # noqa: E501
    r, A = topography.autocorrelation_from_profile(**kwargs)
    return r[1:], np.sqrt(2 * A[1:]) / r[1:]


def scale_dependent_slope_from_area(topography, **kwargs):
    r"""
    Compute the two-dimensional, radially averaged scale-dependent slope.

    The scale-dependent slope is given by

       .. math::
         :nowrap:

         \begin{equation}
         h_\text{rms}^\prime(\lambda) = \left[2A(\lambda)\right]^{1/2}/\lambda
         \end{equation}

    where :math:`A(\lambda)` is the autocorrelation function.

    Parameters
    ----------
    topography : SurfaceTopography or UniformLineScan
        Container storing the uniform topography map
    **kwargs : dict
        Additional keyword parameters are passed on to
        `autocorrelation_from_area`

    Returns
    -------
    r : array
        Distances. (Units: length)
    slope : array
        Slope. (Units: dimensionless)
    """  # noqa: E501
    r, A = topography.autocorrelation_from_area(**kwargs)
    return r[1:], np.sqrt(2 * A[1:]) / r[1:]


# Register analysis functions from this module
UniformTopographyInterface.register_function('scale_dependent_slope_from_profile', scale_dependent_slope_from_profile)
NonuniformLineScanInterface.register_function('scale_dependent_slope_from_profile', scale_dependent_slope_from_profile)
UniformTopographyInterface.register_function('scale_dependent_slope_from_area', scale_dependent_slope_from_area)
NonuniformLineScanInterface.register_function('scale_dependent_slope_from_area', scale_dependent_slope_from_area)



#
# Copyright 2018-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Variable bandwidth analysis for uniform topographies
"""

import numpy as np

from ..Exceptions import NoReliableDataError, UndefinedDataError
from ..HeightContainer import UniformTopographyInterface
from ..Support import build_tuple


def checkerboard_detrend_profile(self, subdivisions, order=1, return_plane=False):
    """
    Perform tilt correction (and substract mean value) in each individual
    line section of a checkerboard decomposition of a profile. For topography
    maps, each horizontal slice is interpreted as a profile and this
    decomposition is carried out for each slice.

    The main application of this function is to carry out a variable
    bandwidth analysis of the surface.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Container storing the uniform topography map
    subdivisions : int
        Number of subdivision, i.e. physical_sizes of the
        checkerboard.
    order : int, optional
        Maximum order of the polynomial used for detrending. (Default: 1)
    return_plane : bool, optional
        Return parameters of the detrending plane. (Default: False)

    Returns
    -------
    arr : np.ndarray
        Array with height information, tilt-corrected within each
        checkerboard.

    if return_plane == True:
    parameters : np.ndarray
        Array of leading order `subdivisions` containing the fit
        parameters.
    """
    # compute unique consecutive index for each subdivided region
    region_index = np.arange(self.nb_grid_pts[0]) * subdivisions // self.nb_grid_pts[0]

    if self.dim == 1:
        x, h = self.positions_and_heights()
    elif self.dim == 2:
        nx, ny = self.nb_grid_pts
        x, y, h = self.positions_and_heights()
        region_index = np.array([[region_index + i * subdivisions] for i in range(ny)]).T
    else:
        raise ValueError(f'Cannot perform checkerboard detrend on topographies of dimension {self.dim}')

    shape = h.shape
    h = h.reshape(-1)
    x = x.reshape(-1)
    region_index = region_index.reshape(-1)

    b = np.array([np.bincount(region_index, h * (x ** i)) for i in range(order + 1)])
    C = np.array([[np.bincount(region_index, x ** (k + i)) for i in range(order + 1)] for k in range(order + 1)])
    a = np.linalg.solve(C.T, b.T).T

    detrended_h = h - np.sum([a[i, region_index] * x ** i for i in range(order + 1)], axis=0)
    detrended_h.shape = shape

    if return_plane:
        return detrended_h, a
    else:
        return detrended_h


def checkerboard_detrend_area(self, subdivisions, order=1, return_plane=False):
    """
    Perform tilt correction (and substract mean value) in each individual
    rectangle of a checkerboard decomposition of the surface. This is
    identical to subdividing the surface into individual, nonoverlapping
    rectangles and performing individual tilt corrections on them.

    The main application of this function is to carry out a variable
    bandwidth analysis of the surface.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Container storing the uniform topography map
    subdivisions : tuple
        Number of subdivision per dimension, i.e. physical_sizes of the
        checkerboard.
    order : int, optional
        Maximum order of the polynomial used for detrending. (Default: 1)
    return_plane : bool, optional
        Return parameters of the detrending plane. (Default: False)

    Returns
    -------
    arr : np.ndarray
        Array with height information, tilt-corrected within each
        checkerboard.

    if return_plane == True:
    parameters : np.ndarray
        Array of leading order `subdivisions` containing the fit
        parameters.
    """
    if self.dim != 2:
        raise ValueError('Areal checkerboard tilt correction can only performend on topography maps, not on profiles.')

    # compute unique consecutive index for each subdivided region
    region_coord = [np.arange(n) * s // n for n, s in zip(self.nb_grid_pts, subdivisions)]
    region_coord = np.meshgrid(*region_coord, indexing='ij')
    region_index = region_coord[0]
    for i in range(1, self.dim):
        region_index = subdivisions[i] * region_index + region_coord[i]

    # Number of polynomial coefficents
    nb_coeff = (order + 1) * (order + 2) // 2

    # Build list of possible exponents
    ij = []
    i = j = 0
    k = nb_coeff
    while k > 0:
        ij += [(i, j)]
        i += 1
        if i + j > order:
            i = 0
            j += 1
        k -= 1

    assert len(ij) == nb_coeff

    x, y, h = self.positions_and_heights()
    shape = h.shape
    h = h.reshape(-1)
    x = x.reshape(-1)
    y = y.reshape(-1)
    region_index = region_index.reshape(-1)

    b = np.array([np.bincount(region_index, h * (x ** i) * (y ** j)) for i, j in ij])
    C = np.array([[np.bincount(region_index, x ** (k + i) * y ** (l + j)) for i, j in ij] for k, l in ij])
    a = np.linalg.solve(C.T, b.T).T

    detrended_h = h - np.sum([a[k, region_index] * (x ** i) * (y ** j) for k, (i, j) in enumerate(ij)], axis=0)
    detrended_h.shape = shape

    if return_plane:
        return detrended_h, a
    else:
        return detrended_h


def variable_bandwidth_from_profile(self, quantities='bh', reliable=True, resampling_method=None, nb_grid_pts_cutoff=4):
    """
    Perform a variable bandwidth analysis by computing the mean
    root-mean-square height within increasingly finer subdivisions of the
    profiles (line scans).

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Container storing the uniform topography map or line scan.
    quantities : str, optional
        Specification of return tuple, where each string character
        stand for specific quantity. Possible quantities are
            - 'm': Magnification (Unit: dimensionless)
            - 'b': Bandwidth (Unit: length)
            - 'h': RMS height (Unit: length)
            - 's': RMS detrending slope (Unit: dimensionless)
        For example, 'mbh' return a tuple with the three entries
        magnification, bandwidth, rms height.
        (Default: 'bh')
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    resampling_method : str, optional
        Can only be None for the variable bandwidth analysis.
        (Default: None)
    nb_grid_pts_cutoff : int, optional
        Minimum nb_grid_pts to allow for subdivision. The analysis will
        automatically analyze subdivision down to this nb_grid_pts.
        (Default: 4)

    Returns
    -------
    magnifications : np.ndarray
        Array containing the magnifications.
    bandwidths : np.ndarray
        Array containing the bandwidths, here the physical sizes of the
        subdivided topography. For 2D topography maps, this is the mean of the
        two physical sizes of the subdivided section of the topography.
    rms_heights : np.ndarray
        Array containing the rms height corresponding to the respective
        magnification.
    rms_slopes : np.ndarray
        Array containing the rms values of the detrending slopes corresponding
        to the respective magnification.
    """
    if resampling_method is not None:
        raise ValueError('`variable_bandwidth_from_profile` does not support resampling.')

    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Variable bandwidth '
                                 'analysis cannot be computed for topographies with missing data points.')

    magnification = 1
    subdivisions = 1
    sx = self.physical_sizes[0]
    nx = int(self.nb_grid_pts[0])
    magnifications = []
    bandwidths = []
    rms_heights = []
    rms_slopes = []
    while nx // subdivisions >= nb_grid_pts_cutoff:
        magnifications += [magnification]
        bandwidths += [sx / subdivisions]
        h, coeffs = self.checkerboard_detrend_profile(subdivisions, return_plane=True)
        rms_heights += [np.sqrt(np.mean(h * h))]
        rms_slopes += [np.sqrt(np.mean(coeffs[1] * coeffs[1]))]
        magnification *= 2
        subdivisions *= 2

    magnifications = np.array(magnifications)
    bandwidths = np.array(bandwidths)
    rms_heights = np.array(rms_heights)
    rms_slopes = np.array(rms_slopes)

    if reliable:
        short_cutoff = self.short_reliability_cutoff()
        if short_cutoff:
            m = bandwidths > short_cutoff
            if m.sum() == 0:
                raise NoReliableDataError('Dataset contains no reliable data.')
            magnifications = magnifications[m]
            bandwidths = bandwidths[m]
            rms_heights = rms_heights[m]
            rms_slopes = rms_slopes[m]

    return build_tuple(quantities, m=magnifications, b=bandwidths, h=rms_heights, g=rms_slopes)


def variable_bandwidth_from_area(self, quantities='bh', reliable=True, resampling_method=None, nb_grid_pts_cutoff=4):
    """
    Perform a variable bandwidth analysis by computing the mean
    root-mean-square height within increasingly finer subdivisions of the
    surface topography.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Container storing the uniform topography map.
    quantities : str, optional
        Specification of return tuple, where each string character
        stand for specific quantity. Possible quantities are
            - 'm': Magnification (Unit: dimensionless)
            - 'b': Bandwidth (Unit: length)
            - 'h': RMS height (Unit: length)
            - 'g': RMS detrending gradient (Unit: dimensionless)
        For example, 'mbh' return a tuple with the three entries
        magnification, bandwidth, rms height.
        (Default: 'bh')
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    resampling_method : str, optional
        Can only be None for the variable bandwidth analysis.
        (Default: None)
    nb_grid_pts_cutoff : int, optional
        Minimum nb_grid_pts to allow for subdivision. The analysis will
        automatically analyze subdivision down to this nb_grid_pts.
        (Default: 4)

    Returns
    -------
    magnifications : array
        Array containing the magnifications.
    bandwidths : array
        Array containing the bandwidths, here the physical sizes of the
        subdivided topography. For 2D topography maps, this is the mean of the
        two physical sizes of the subdivided section of the topography.
    rms_heights : array
        Array containing the rms height corresponding to the respective
        magnification.
    rms_gradients : np.ndarray
        Array containing the rms values of the detrending gradients
        corresponding to the respective magnification.
    """
    if resampling_method is not None:
        raise ValueError('`variable_bandwidth_from_profile` does not support resampling.')

    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Variable bandwidth '
                                 'analysis cannot be computed for topographies with missing data points.')

    magnification = 1
    physical_sizes = np.array(self.physical_sizes)
    min_size = np.min(physical_sizes)
    subdivisions = np.round(self.physical_sizes / min_size).astype(int)
    nb_grid_pts = np.array(self.nb_grid_pts, dtype=int)
    magnifications = []
    bandwidths = []
    rms_heights = []
    rms_gradients = []
    while ((nb_grid_pts // subdivisions).min() >= nb_grid_pts_cutoff):
        magnifications += [magnification]
        bandwidths += [np.mean(physical_sizes / subdivisions)]
        h, coeffs = self.checkerboard_detrend_area(subdivisions, return_plane=True)
        rms_heights += [np.sqrt(np.mean(h * h))]
        rms_gradients += [np.sqrt(np.mean(coeffs[1] * coeffs[1] + coeffs[2] * coeffs[2]))]
        magnification *= 2
        subdivisions *= 2

    magnifications = np.array(magnifications)
    bandwidths = np.array(bandwidths)
    rms_heights = np.array(rms_heights)
    rms_gradients = np.array(rms_gradients)

    if reliable:
        short_cutoff = self.short_reliability_cutoff()
        if short_cutoff:
            m = bandwidths > short_cutoff
            if m.sum() == 0:
                raise NoReliableDataError('Dataset contains no reliable data.')
            magnifications = magnifications[m]
            bandwidths = bandwidths[m]
            rms_heights = rms_heights[m]
            rms_gradients = rms_gradients[m]

    return build_tuple(quantities, m=magnifications, b=bandwidths, h=rms_heights, g=rms_gradients)


# Register analysis functions from this module
UniformTopographyInterface.register_function('checkerboard_detrend_profile', checkerboard_detrend_profile)
UniformTopographyInterface.register_function('checkerboard_detrend_area', checkerboard_detrend_area)
UniformTopographyInterface.register_function('variable_bandwidth_from_profile', variable_bandwidth_from_profile)
UniformTopographyInterface.register_function('variable_bandwidth_from_area', variable_bandwidth_from_area)



#
# Copyright 2020 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import os
import sys
from collections import defaultdict
from datetime import datetime
from subprocess import Popen, PIPE

root = os.path.dirname(sys.argv[0])


def read_authors(fn):
    return {email.strip('<>'): name for name, email in
            [line.rsplit(maxsplit=1) for line in open(fn, 'r')]}


def parse_git_log(log, authors):
    committers = defaultdict(set)
    author = None
    date = None
    for line in log.decode('utf-8').split('\n'):
        if line.startswith('commit'):
            if date is not None and author is not None:
                committers[author].add(date.year)
        elif line.startswith('Author:'):
            email = line.rsplit('<', maxsplit=1)[1][:-1]
        elif line.startswith('Date:'):
            date = datetime.strptime(line[5:].rsplit(maxsplit=1)[0].strip(),
                                     '%a %b %d %H:%M:%S %Y')
            try:
                author = authors[email]
            except KeyError:
                author = email
        elif 'copyright' in line.lower() or 'license' in line.lower():
            date = None
    if date is not None:
        committers[author].add(date.year)
    return committers


def pretty_years(years):
    def add_to_year_string(s, pprev_year, prev_year):
        if pprev_year == prev_year:
            # It is a single year
            if s is None:
                return f'{prev_year}'
            else:
                return f'{s}, {prev_year}'
        else:
            # It is a range
            if s is None:
                return f'{pprev_year}-{prev_year}'
            else:
                return f'{s}, {pprev_year}-{prev_year}'

    years = sorted(years)
    prev_year = pprev_year = years[0]
    s = None
    for year in years[1:]:
        if year - prev_year > 1:
            s = add_to_year_string(s, pprev_year, prev_year)
            pprev_year = year
        prev_year = year
    return add_to_year_string(s, pprev_year, prev_year)


authors = read_authors('{}/../AUTHORS'.format(root))

process = Popen(['git', 'log', '--follow', sys.argv[1]], stdout=PIPE,
                stderr=PIPE)
stdout, stderr = process.communicate()
committers = parse_git_log(stdout, authors)

prefix = 'Copyright'
for name, years in committers.items():
    print('{} {} {}'.format(prefix, pretty_years(years), name))
    prefix = ' ' * len(prefix)
print()



#
# Copyright 2019-2021 Lars Pastewka
#           2020-2021 Michael Röttger
#           2019-2020 Kai Haase
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from ..Exceptions import CorruptFile, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography
from ..Support.UnitConversion import mangle_length_unit_utf8

from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo

image_head = b'fileType      Image\n'
spec_head = b'fileType      Spectroscopy\n'

magic_data = b'data          \n'
magic_data_binary = b'data          BINARY\n'
magic_data_binary32 = b'data          BINARY_32\n'
magic_data_ascii = b'data          ASCII'


class MIReader(ReaderBase):
    _format = 'mi'
    _name = 'Molecular Imaging (MI)'
    _description = '''
Agilent Technologies (Molecular Imaging) AFM saves filed in the MI format.
This format contains information on the physical size of the topography map as
well as its units.
'''

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self.file_path = file_path

        # Start of the data in the file
        self.data_start = None
        # If image oder spectroscopy
        self.is_image = None
        # Format of the saved data
        self.data_type = None

        self.mifile = None

        # Process the header and read in metadata
        self.process_header()

    def process_header(self):
        with OpenFromAny(self.file_path, 'rb') as f:
            self.lines = f.readlines()

            # Find out if image or spectroscopy
            if self.lines[0] == image_head:
                self.is_image = True
            elif self.lines[0] == spec_head:
                self.is_image = False
            else:
                raise CorruptFile

            # Find the start of the height data and denote data type
            if magic_data in self.lines:
                header_size = self.lines.index(magic_data)
                self.data_type = 'text'
            elif magic_data_binary in self.lines:
                header_size = self.lines.index(magic_data_binary)
                self.data_type = 'binary'
            elif magic_data_binary32 in self.lines:
                header_size = self.lines.index(magic_data_binary32)
                self.data_type = 'binary32'
            elif magic_data_ascii in self.lines:
                header_size = self.lines.index(magic_data_ascii)
                self.data_type = 'ascii'
            else:
                raise CorruptFile

            # Save start of data for later reading of the matrix
            self.data_start = header_size + 1

            # Create mifile from header, reading out meta data and channel info
            if self.is_image:
                self.mifile = read_header_image(self.lines[1:header_size])
            else:  # TODO
                self.mifile = read_header_spect(self.lines[1:header_size])

            # Reformat the metadata
            for buf in self.mifile.channels:
                buf.meta['name'] = buf.name
                buf.unit = mangle_length_unit_utf8(buf.meta.pop('bufferUnit'))
                buf.meta['range'] = buf.meta.pop('bufferRange')
                buf.meta['label'] = buf.meta.pop('bufferLabel')

            self._physical_sizes = float(self.mifile.meta['xLength']), float(self.mifile.meta['yLength'])
            self._nb_grid_pts = int(self.mifile.meta['xPixels']), int(self.mifile.meta['yPixels'])

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={}, periodic=False,
                   subdomain_locations=None, nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')

        output_channel = self.mifile.channels[channel_index]

        buffer = b''.join(self.lines[self.data_start:])
        buffer = [chr(_) for _ in buffer]

        # Read height data
        if self.is_image:
            if self.data_type == 'binary':
                dt = "i2"
                encode_length = 2
                type_range = 32768
            elif self.data_type == 'binary32':
                dt = "i4"
                encode_length = 4
                type_range = 2147483648
            else:  # text or ascii
                type_range = 32768

            start = int(self.mifile.xres) * int(
                self.mifile.yres) * encode_length * channel_index
            end = int(self.mifile.xres) * int(
                self.mifile.yres) * encode_length * (channel_index + 1)

            data = ''.join(buffer[start:end])
            out = np.frombuffer(str.encode(data, "raw_unicode_escape"), dt)
            out = out.reshape((int(self.mifile.xres), int(self.mifile.yres)))

            # Undo normalizing with range of data type
            out = out / type_range

            # If scan direction is upwards, flip the height map vertically
            if self.mifile.meta['scanUp']:
                out = np.flip(out, 0)

            # Multiply the heights with the bufferRange
            out *= float(output_channel.meta['range'])
        else:
            pass  # TODO

        joined_meta = {**self.mifile.meta, **output_channel.meta}
        info = info.copy()
        info.update(joined_meta)

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        # Initialize heights with transposed array in order to match Gwdyydion
        # when plotted with pcolormesh(t.heights().T), except that the y axis
        # is flipped because the origin is in lower left with pcolormesh;
        # imshow(t.heights().T) shows the image like gwyddion
        t = Topography(heights=out.T,
                       physical_sizes=self._check_physical_sizes(
                           physical_sizes, self._physical_sizes),
                       unit=output_channel.unit, info=info, periodic=periodic)
        if height_scale_factor is not None:
            t = t.scale(height_scale_factor)
        return t

    @property
    def channels(self):
        return [ChannelInfo(self, i, name=channel.meta['name'],
                            dim=len(self._nb_grid_pts),
                            nb_grid_pts=self._nb_grid_pts,
                            physical_sizes=self._physical_sizes,
                            uniform=True,
                            unit=channel.unit,
                            info=channel.meta)
                for i, channel in enumerate(self.mifile.channels)]

    @property
    def info(self):
        """Return all the available metadata as a dict."""
        return self.mifile.meta

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__


def read_header_image(header):
    """
    Reads in global metadata and information about about included channels as
    well as their metadata.

    Parameters
    ----------
    header : str
        The header as a line of text.

    Returns
    -------
    mifile : MIFile
        Data structure item containing the metadata, with the channels as a
        list of 'buffers'.
    """
    # This object will store the file-wide metadata
    mifile = MIFile()

    # False while reading global metadata, gets true if we start to read in the
    # channel info
    reading_buffers = False

    for line in header:
        line = line.decode("utf-8")

        # As soon as we see a line starting with 'bufferLabel', we know we are
        # now reading in channels
        if line.startswith('bufferLabel'):
            # Create a new channel with the id as name
            channel = Channel(name=str.strip(line[14:]))
            mifile.channels.append(channel)
            reading_buffers = True

        if line[0] == ' ':
            continue

        # For all key value pairs in the file:
        # Append to global metadata oder channel metadata, depending on our
        # state
        key = str.strip(line[:14])
        value = str.strip(line[14:])

        if reading_buffers:
            mifile.channels[-1].meta[key] = value
        else:
            mifile.meta[key] = value

        # Catch 'special' metadata
        if key == "xPixels":
            mifile.xres = value
        elif key == "yPixels":
            mifile.yres = value
    return mifile


def read_header_spect(header):
    """
    Reads in metadata out of the header.

    Parameters
    ----------
    header : str
        The header.

    Returns
    -------
    mifile : MIFile
        Data structure item containing the metadata.
    """
    raise NotImplementedError


class Channel:
    """
    Class structure for a channel contained in the file.
    Has a name and metadata (height data is not needed since it is returned
    directly).
    """

    def __init__(self, name=None, unit=None, meta=None):
        if meta is None:
            meta = dict()

        self.name = name
        self.unit = unit
        self.meta = meta


class MIFile:
    """
    Class structure for the while file. Has a list of channels, global metadata
    and a nb_grid_pts.
    """

    def __init__(self, res=(0, 0), channels=None, meta=None):
        if channels is None:
            channels = list()
        if meta is None:
            meta = dict()

        self.xres = res[0]
        self.yres = res[1]
        self.channels = channels
        self.meta = meta



#
# Copyright 2019-2021 Lars Pastewka
#           2019-2020 Antoine Sanner
#           2019 Michael Röttger
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Base class for geometric topogography descriptions
"""

import abc

import numpy as np

from NuMPI import MPI
from NuMPI.Tools import Reduction

from .Support import doi, DeprecatedDictionary

# Standardized entries for the info dictionary

# Length unit of the measurement. The unit applies to both lateral and heights.
# Data type: str
INFO_UNIT = 'unit'
# Data and time of the data acquisition.
# Data type: datetime.datetime
INFO_ACQUISITION_TIME = 'acquisition_time'


class AbstractTopography(object):
    """
    Base class for all classes storing height information.

    The member dictionary `_functions` contains a list of functions that
    can be executed on this specific class.

    The dictionary itself is owned by the interface,
    `UniformTopographyInterface` and `NonuniformLineScanInterface`. This is
    because the functions are determined by the type of topography that is
    represented, not by the pipeline hierarchy. For example, converters that
    convert uniform to nonuniform and vice versa need to have the respective
    interface of the format they are converting to.
    """

    class Error(Exception):
        # pylint: disable=missing-docstring
        pass

    def __init__(self, unit=None, info={}, communicator=MPI.COMM_WORLD):
        self._unit = unit
        self._info = info.copy()
        self._communicator = communicator

    def apply(self, name, *args, **kwargs):
        self._functions[name](self, *args, **kwargs)

    def __getattr__(self, name):
        if name in self._functions:
            def func(*args, **kwargs):
                return self._functions[name](self, *args, **kwargs)

            func.__doc__ = self._functions[name].__doc__
            return func
        else:
            raise AttributeError(
                "Unkown attribute '{}' and no analysis or pipeline function of this name registered (class {}). "
                "Available functions: {}".format(name, self.__class__.__name__, ', '.join(self._functions.keys())))

    def __dir__(self):
        return sorted(super().__dir__() + [*self._functions])

    def __getstate__(self):
        """
        Upon pickling, it is called and the returned object is pickled as the
        contents for the instance.
        """
        return self._unit, self._info

    def __setstate__(self, state):
        """
        Upon unpickling, it is called with the unpickled state.
        The argument `state` is the result of `__getstate__`.
        """
        self._unit, self._info = state
        self._communicator = MPI.COMM_WORLD

    @property
    @abc.abstractmethod
    def is_periodic(self):
        """Return whether the topography is periodically repeated at the
        boundaries."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dim(self):
        """Returns 1 for line scans and 2 for topography maps."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def physical_sizes(self, ):
        """Return the physical sizes of the topography."""
        raise NotImplementedError

    @property
    def unit(self):
        """Return the length unit of the topography."""
        return self._unit

    @property
    def info(self):
        """
        Return the info dictionary. The info dictionary contains auxiliary data
        found in the topography data file but not directly used by PyCo.

        The dictionary can contain any type of information. There are a few
        standardized keys, listed in the following.

        Standardized keys
        -----------------
        'acquisition_time' : :obj:`datetime`
            Date and time of the measurement.
        'instrument' : dict
            Nested dictionary containing instrument data.
                - 'name' : str - Name of instrument
                - 'parameters' : dict - Additional instrument parameters, e.g.
                    - 'tip_radius' : dict - Tip radius for tip based techniques
                        - 'value' : float - Value of tip radius
                        - 'unit' : str - Unit of tip radius
        'raw_metadata' : dict
            Dictionary containing the raw (non-standardized) metadata as
            extracted during reading of the topography.
        """
        info = DeprecatedDictionary(self._info, deprecated_keys=['unit'])
        if self.unit is not None:
            info.update(dict(unit=self.unit))
        return info

    @property
    def communicator(self):
        """Return the MPI communicator object."""
        return self._communicator

    def pipeline(self):
        return [self]


class DecoratedTopography(AbstractTopography):
    """
    Base class of topographies with parent. Having a parent means that the
    data is owned by the parent, but the present class performs
    transformations on that data. This is a simple realization of a
    processing pipeline. Note that child topographies don't store their
    own physical_sizes etc. but pass this information through to the parent.
    """

    def __init__(self, topography, unit=None, info={}):
        """
        Arguments
        ---------
        topography : SurfaceTopography
            The parent topography.
        """
        super().__init__(unit=unit, info=info)
        assert isinstance(topography, AbstractTopography)
        self.parent_topography = topography
        self._communicator = self.parent_topography.communicator

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self.parent_topography
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self.parent_topography = state
        super().__setstate__(superstate)

    @property
    def info(self):
        """ Return info dictionary """
        info = self.parent_topography._info.copy()
        info.update(self._info)
        return info

    @property
    def nb_subdomain_grid_pts(self):
        return self.parent_topography.nb_subdomain_grid_pts

    def pipeline(self):
        return self.parent_topography.pipeline() + [self]


class TopographyInterface(object):
    @classmethod
    def register_function(cls, name, function):
        if function.__name__ != 'func_with_doi':
            # We want the `dois` argument for all pipeline functions. If no
            # doi has been specified, we simply wrap it in an empty decorator.
            cls._functions.update({name: doi()(function)})
        else:
            cls._functions.update({name: function})


class UniformTopographyInterface(TopographyInterface, metaclass=abc.ABCMeta):
    _functions = {}

    @property
    def is_uniform(self):
        return True

    @property
    def is_reentrant(self):
        return False  # Uniform datasets cannot be reentrant

    @property
    def is_domain_decomposed(self):
        return self.nb_grid_pts != self.nb_subdomain_grid_pts

    @property
    def communicator(self):
        # default value is COMM_SELF because sometimes NON-MPI readers that
        # do not set the communicator value are used in MPI Programs.
        # See discussion in issue #166
        return MPI.COMM_SELF

    @property
    @abc.abstractmethod
    def nb_grid_pts(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def nb_subdomain_grid_pts(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def pixel_size(self):
        return NotImplementedError

    @property
    @abc.abstractmethod
    def area_per_pt(self):
        return NotImplementedError

    @property
    @abc.abstractmethod
    def has_undefined_data(self):
        return NotImplementedError

    @abc.abstractmethod
    def positions(self):
        """
        Returns array containing the lateral positions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def heights(self):
        """
        Returns array containing the topography data.
        """
        return NotImplementedError

    def positions_and_heights(self, **kwargs):
        """
        Returns array containing the lateral positions and the topography
        data.
        """
        p = self.positions(**kwargs)
        h = self.heights()
        try:
            x, y = p
            return x, y, h
        except ValueError:
            return p, h

    def __eq__(self, other):
        return Reduction(self._communicator).all(
            self.unit == other.unit and self.info == other.info and self.is_periodic == other.is_periodic and
            np.allclose(self.positions(), other.positions()) and np.allclose(self.heights(), other.heights()))

    def __getitem__(self, i):
        return self.heights()[i]


class NonuniformLineScanInterface(TopographyInterface, metaclass=abc.ABCMeta):
    _functions = {}

    @property
    def is_uniform(self):
        return False

    @property
    def is_reentrant(self):
        return np.min(np.diff(self.positions())) <= 0

    @property
    @abc.abstractmethod
    def nb_grid_pts(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def x_range(self):
        raise NotImplementedError

    @abc.abstractmethod
    def positions(self):
        """
        Returns array containing the lateral positions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def heights(self):
        """
        Returns array containing the topography data.
        """
        raise NotImplementedError

    def positions_and_heights(self, **kwargs):
        """
        Returns array containing the lateral positions and the topography
        data.
        """
        return self.positions(**kwargs), self.heights()

    @property
    def is_MPI(self):
        return False

    @property
    def has_undefined_data(self):
        return False

    def __eq__(self, other):
        return Reduction(self._communicator).all(
            self.unit == other.unit and self.info == other.info and self.is_periodic == other.is_periodic and
            np.allclose(self.positions_and_heights(), other.positions_and_heights()))

    def __getitem__(self, i):
        return self.positions()[i], self.heights()[i]
